# Generated by Django 2.2.2 on 2019-06-26 16:20

from django.db import migrations


def forward_filter_rename(apps, schema_editor):
    AuthToken = apps.get_model("spider_base", "AuthToken")

    for i in AuthToken.objects.all():
        i.extra["search"] = i.extra.pop("filter", None) or []
        i.extra["request_search"] = i.extra.pop("request_filter", None)
        i.save(update_fields=["extra"])


def backward_filter_rename(apps, schema_editor):
    AuthToken = apps.get_model("spider_base", "AuthToken")

    for i in AuthToken.objects.all():
        i.extra["filter"] = i.extra.pop("search", None) or []
        i.extra["request_filter"] = i.extra.pop("request_search", None)
        i.save(update_fields=["extra"])


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0004_auto_20190621_2126'),
    ]

    operations = [
        migrations.RunPython(forward_filter_rename, backward_filter_rename),
    ]




import logging
import random
import os
import time

from django.contrib.auth.backends import ModelBackend
from django.http import Http404

import ratelimit

from .models import UserComponent, Protection, TravelProtection
from .constants import (
    ProtectionType, MIN_PROTECTION_STRENGTH_LOGIN,
)

logger = logging.getLogger(__name__)

# seed with real random
_nonexhaustRandom = random.Random(os.urandom(30))


class SpiderAuthBackend(ModelBackend):

    def authenticate(self, request, username=None,
                     protection_codes=None, nospider=False, **kwargs):
        """ Use protections for authentication"""
        # disable SpiderAuthBackend backend (against recursion)
        if nospider:
            return
        uc = UserComponent.objects.filter(
            user__username=username, name="index"
        ).first()
        if not uc:
            request.protections = Protection.authall(
                request, scope="auth",
                ptype=ProtectionType.authentication.value,
                protection_codes=protection_codes
            )
            if type(request.protections) is int:  # should never happen
                logger.warning(
                    "Login try without username, should never "
                    "happen, archieved strength: %s",
                    request.protections
                )
                return None
        else:
            try:
                request.protections = uc.auth(
                    request, scope="auth",
                    ptype=ProtectionType.authentication.value,
                    protection_codes=protection_codes
                )
            except Http404:
                # for Http404 auth abort by protections (e.g. Random Fail)
                pass

            if type(request.protections) is int:
                if TravelProtection.objects.auth(request, uc):
                    if request.protections < MIN_PROTECTION_STRENGTH_LOGIN:
                        logger.warning(
                            "Low login protection strength: %s, %s",
                            request.protections, username
                        )
                    return uc.user
        # error path

        # allow blocking per hour
        ratelimit.get_ratelimit(
            request=request, group="spider_login_failed_ip", key="ip",
            inc=True, rate=(float("inf"), 3600)
        )
        ratelimit.get_ratelimit(
            request=request, group="spider_login_failed_account",
            key=lambda x, y: username, inc=True, rate=(float("inf"), 3600)
        )
        # be less secure here, most probably the user is already known
        time.sleep(_nonexhaustRandom.random()/2)



# Generated by Django 2.1 on 2018-08-20 14:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0001_initial'),
        ('spider_filets', '0002_textfilet_non_public_edit'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='textfilet',
            name='non_public_edit',
        ),
        migrations.AddField(
            model_name='textfilet',
            name='editable_from',
            field=models.ManyToManyField(help_text='Allow editing from selected componentsby privileged users.', related_name='_textfilet_editable_from_+', to='spider_base.UserComponent'),
        ),
    ]



__all__ = [
    "SpiderAuthForm"
]


from django import forms
from django.contrib.auth.forms import AuthenticationForm

from ..auth import SpiderAuthBackend
from ..constants import ProtectionType

from ..models import (
    Protection
)


class SpiderAuthForm(AuthenticationForm):
    password = None
    # can authenticate only with specialized backend(s) so ignore others
    auth_backend = SpiderAuthBackend()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.is_bound:
            self.reset_protections()

    @property
    def media(self):
        return super().media + self.request.protections.media

    def reset_protections(self):
        self.request.protections = Protection.authall(
            self.request, scope="auth",
            ptype=ProtectionType.authentication.value,
        )
        # here is not even a usercomponent available
        # if this works, something is wrong
        # protections should match username from POST with the ones of the
        # usercomponent (which is available in clean)
        assert type(self.request.protections) is not int,\
            "login evaluates to success without data"
        assert len(self.request.protections) > 0

    def clean(self):
        username = self.cleaned_data.get('username')
        protection_codes = None
        if "protection" in self.request.GET:
            protection_codes = self.request.GET.getlist("protection")

        if username is not None:
            self.user_cache = self.auth_backend.authenticate(
                self.request, username=username,
                protection_codes=protection_codes
            )
            if self.user_cache is None:
                raise forms.ValidationError(
                    self.error_messages['invalid_login'],
                    code='invalid_login',
                    params={'username': self.username_field.verbose_name},
                )
            else:
                self.confirm_login_allowed(self.user_cache)
        return self.cleaned_data



# Generated by Django 2.2 on 2019-04-17 19:28

from django.db import migrations, models
import spkcspider.apps.spider_accounts.models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_accounts', '0004_auto_20190109_1505'),
    ]

    operations = [
        migrations.AddField(
            model_name='spideruser',
            name='quota_usercomponents',
            field=models.PositiveIntegerField(blank=True, default=spkcspider.apps.spider_accounts.models.default_quota_spider_user_components, help_text='Quota in units, null for no limit', null=True),
        ),
    ]



__all__ = ["literalize", "uriref", "spkc_namespace"]

import logging

from django import template

from rdflib import Literal
from rdflib.namespace import XSD, URIRef
from django.forms import BoundField

from ..constants import spkcgraph
from ..helpers import (
    merge_get_url,
    literalize as _literalize,
    field_to_python as _field_to_python
)

register = template.Library()


@register.filter()
def uriref(path):
    return URIRef(path)


@register.filter()
def is_uriref(value):
    return isinstance(value, URIRef)


@register.filter()
def field_to_python(value):
    return _field_to_python(value)


@register.simple_tag(takes_context=True)
def action_view(context):
    token = getattr(context["request"], "auth_token", None)
    if token:
        token = token.token
    url2 = merge_get_url(
        context["hostpart"] + context["request"].path, token=token
    )
    return Literal(url2, datatype=XSD.anyURI)


@register.simple_tag(takes_context=True)
def literalize(
    context, ob, datatype=None, use_uriref=None, domain_base=None
):
    if domain_base is None:
        if "hostpart" not in context:
            logging.warning(
                "%s has neither hostpart nor explicit domain_base",
                context["request"].get_full_path()
            )
            context["hostpart"] = "{}://{}".format(
                context["request"].scheme, context["request"].get_host()
            )
        domain_base = context["hostpart"]
    return _literalize(
        ob,
        datatype=datatype, use_uriref=use_uriref, domain_base=domain_base
    )


@register.simple_tag()
def hashable_literalize(field):
    if isinstance(field, BoundField):
        return _literalize(getattr(field.field, "hashable", False))
    else:
        return _literalize(getattr(field, "hashable", False))


@register.simple_tag()
def spkc_namespace(sub=None):
    if sub:
        return spkcgraph[sub]
    return spkcgraph



"""
Base for Contents
namespace: spider_base

"""

__all__ = [
    "ContentVariant", "AssignedContent"
]

import logging

from django.db import models
from django.conf import settings
from django.utils.translation import gettext, gettext_lazy as _
from django.urls import reverse
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.core import validators
from django.apps import apps

from ..contents import installed_contents

# from ..constants import VariantType
from ..helpers import create_b64_id_token
from ..constants import (
    MAX_TOKEN_B64_SIZE, VariantType, hex_size_of_bigid, static_token_matcher
)
from ..validators import content_name_validator, validator_token

from .base import BaseInfoModel, BaseSubUserComponentModel

logger = logging.getLogger(__name__)


# ContentType is already occupied
class ContentVariant(models.Model):
    id = models.BigAutoField(primary_key=True, editable=False)
    # usercontent abilities/requirements
    ctype = models.CharField(
        max_length=10, default="",
    )
    code = models.CharField(max_length=255)
    name = models.SlugField(max_length=50, unique=True, db_index=True)
    # required protection strength (selection)
    strength = models.PositiveSmallIntegerField(
        default=0, validators=[validators.MaxValueValidator(10)]
    )
    # for content features
    valid_feature_for = models.ManyToManyField(
        "self", blank=True, related_name="valid_features", symmetrical=False
    )

    @property
    def installed_class(self):
        return installed_contents[self.code]

    @property
    def feature_urls(self):
        return installed_contents[self.code].cached_feature_urls(self.name)

    def localize_name(self):
        if self.code not in installed_contents:
            return self.name
        return self.installed_class.localize_name(self.name)

    @property
    def unique_for_component(self):
        return VariantType.unique.value in self.ctype

    def __str__(self):
        return self.localize_name()

    def __repr__(self):
        return "<ContentVariant: %s>" % self.__str__()


class UserContentManager(models.Manager):

    def create(self, **kwargs):
        ret = self.get_queryset().create(**kwargs)
        if not ret.token:
            ret.token = create_b64_id_token(ret.id, "_")
            ret.save(update_fields=["token"])
        return ret

    def update_or_create(self, **kwargs):
        ret = self.get_queryset().update_or_create(**kwargs)
        if not ret[0].token:
            ret[0].token = create_b64_id_token(ret[0].id, "_")
            ret[0].save(update_fields=["token"])
        return ret

    def get_or_create(self, defaults=None, **kwargs):
        ret = self.get_queryset().get_or_create(**kwargs)
        if not ret[0].token:
            ret[0].token = create_b64_id_token(ret[0].id, "_")
            ret[0].save(update_fields=["token"])
        return ret

    def from_token(
        self, token, *, info=frozenset(), variant=frozenset(),
        check_feature=False
    ):
        # check_feature requires that ctype of the selected content is in
        #  component features
        if isinstance(variant, str):
            variant = [variant]

        if isinstance(info, str):
            info = [info]

        q = models.Q()
        if variant:
            if isinstance(variant, models.QuerySet):
                q &= models.Q(ctype__in=variant)
            elif isinstance(variant, models.Model):
                q &= models.Q(ctype=variant)
            else:
                q &= models.Q(ctype__name__in=variant)
        for i in info:
            q &= models.Q(info__contains="\x1e%s\x1e" % i)
        if check_feature:
            q &= models.Q(
                ctype__feature_for_components__authtokens=token
            )
        return self.get(
            q, attached_to_token=token
        )

    def from_url_part(
        self, url, *, info=frozenset(), variant=frozenset(),
        check_feature=False
    ):
        """
            url: can be full url or token/accessmethod
            info: should be either string or iterable which will be matched
                  against info to retrieve an unique content
            returns: (<matched content>, <current content>/None)
        """
        UserComponent = apps.get_model("spider_base", "UserComponent")
        res = static_token_matcher.match(url)
        if not res:
            raise self.model.DoesNotExist()
        res = res.groupdict()
        # shortcut, we don't need to look up db to see that without
        # info field multiple items can match easily
        # the special case: only one content is unreliable and should
        # never tried to match
        if not info and not variant:
            raise self.model.MultipleObjectsReturned()

        if isinstance(variant, str):
            variant = [variant]

        if isinstance(info, str):
            info = [info]

        q = models.Q()
        if variant:
            if isinstance(variant, models.QuerySet):
                q &= models.Q(ctype__in=variant)
            elif isinstance(variant, models.Model):
                q &= models.Q(ctype=variant)
            else:
                q &= models.Q(ctype__name__in=variant)
        for i in info:
            q &= models.Q(info__contains="\x1e%s\x1e" % i)
        if res["access"] == "list":
            uc = UserComponent.objects.filter(
                token=res["static_token"]
            ).first()
            if check_feature:
                q &= (
                    models.Q(
                        ctype__feature_for_components=uc
                    )
                )
            return (self.get(q, usercomponent=uc), None)
        else:
            content = self.get(
                token=res["static_token"]
            )
            if check_feature:
                q &= (
                    models.Q(
                        ctype__feature_for_contents=content
                    ) |
                    models.Q(
                        ctype__feature_for_components=content.usercomponent
                    )
                )
            return (
                self.get(
                    q, usercomponent=content.usercomponent
                ),
                content
            )


class AssignedContent(BaseInfoModel, BaseSubUserComponentModel):
    id = models.BigAutoField(primary_key=True, editable=False)
    # fake_level = models.PositiveIntegerField(null=False, default=0)
    attached_to_token = models.ForeignKey(
        "spider_base.AuthToken", blank=True, null=True,
        on_delete=models.CASCADE
    )
    # don't use extensive recursion,
    # this can cause performance problems and headaches
    # this is not enforced for allowing some small chains
    # see SPIDER_MAX_EMBED_DEPTH setting for limits (default around 5)
    attached_to_content = models.ForeignKey(
        "self", blank=True, null=True,
        related_name="attached_contents", on_delete=models.CASCADE
    )
    # set to indicate creating a new token
    token_generate_new_size = None
    # brute force protection and identifier, replaces nonce
    #  16 = usercomponent.id in hexadecimal
    #  +1 for seperator
    token = models.CharField(
        max_length=(MAX_TOKEN_B64_SIZE)+hex_size_of_bigid+2,
        db_index=True, unique=True, null=True, blank=True,
        validators=[validator_token]
    )
    # regex disables controlcars and disable special spaces
    # and allows some of special characters
    name = models.CharField(
        max_length=255, blank=True, default="",
        validators=[
            content_name_validator
        ]
    )
    description = models.TextField(
        default="", blank=True
    )
    usercomponent = models.ForeignKey(
        "spider_base.UserComponent", on_delete=models.CASCADE,
        related_name="contents", null=False, blank=False
    )
    features = models.ManyToManyField(
        "spider_base.ContentVariant",
        related_name="feature_for_contents", blank=True,
        limit_choices_to=models.Q(
            ctype__contains=VariantType.content_feature.value
        )
    )
    # ctype is here extended: VariantObject with abilities, name, model_name
    ctype = models.ForeignKey(
        ContentVariant, editable=False, null=True,
        on_delete=models.SET_NULL
    )
    priority = models.SmallIntegerField(default=0, blank=True)

    # creator = models.ForeignKey(
    #    settings.AUTH_USER_MODEL, editable=False, null=True,
    #    on_delete=models.SET_NULL
    # )
    created = models.DateTimeField(auto_now_add=True, editable=False)
    modified = models.DateTimeField(auto_now=True, editable=False)
    # only editable for admins
    deletion_requested = models.DateTimeField(
        null=True, blank=True, default=None
    )
    # required protection strength (real)
    strength = models.PositiveSmallIntegerField(
        default=0, validators=[validators.MaxValueValidator(10)],
        editable=False
    )
    # required protection strength for links, 11 to disable links
    strength_link = models.PositiveSmallIntegerField(
        default=0, validators=[validators.MaxValueValidator(11)],
        editable=False
    )
    attached_to_primary_anchor = models.BooleanField(
        default=False, editable=False, null=False,
        help_text=_(
            "Content references primary anchor"
        )
    )
    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, editable=False,
        related_name="+"
    )
    object_id = models.BigIntegerField(editable=False)
    content = GenericForeignKey(
        'content_type', 'object_id', for_concrete_model=False
    )
    # for quick retrieval!! even maybe a duplicate
    # layouts referencing models are not appearing here, so do it here
    references = models.ManyToManyField(
        "self", related_name="referenced_by", editable=False,
        symmetrical=False
    )
    # info extra flags:
    #  primary: primary content of type for usercomponent
    #  unlisted: not listed
    objects = UserContentManager()

    class Meta:
        unique_together = [
            ('content_type', 'object_id'),
        ]
        constraints = []
        if (
            settings.DATABASES["default"]["ENGINE"] !=
            "django.db.backends.mysql"
        ):
            constraints.append(
                models.UniqueConstraint(
                    fields=['usercomponent', 'info'], name='unique_info'
                )
            )

    def __init__(self, *args, **kwargs):
        if isinstance(kwargs.get("ctype"), str):
            kwargs["ctype"] = ContentVariant.objects.get(name=kwargs["ctype"])
        super().__init__(*args, **kwargs)

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<AssignedContent: ({}: {})>".format(
            self.usercomponent.username,
            self.name
        )

    def get_absolute_url(self, scope="view"):
        return reverse(
            "spider_base:ucontent-access",
            kwargs={"token": self.token, "access": scope}
        )

    # use next_object, last_object instead get_next_by_FOO, ...
    # for preventing disclosure of elements

    def get_size(self):
        if not self.content:
            return 0
        return self.content.get_size()

    def localized_description(self):
        """ """
        if not self.content:
            return self.description
        return self.content.localized_description()

    def clean(self):
        _ = gettext
        if VariantType.persist.value in self.ctype.ctype:
            if not self.attached_to_token:
                raise ValidationError(
                    _('Persistent token required'),
                    code="persist",
                )

        if not self.usercomponent.user_info.allowed_content.filter(
            name=self.ctype.name
        ).exists():
            raise ValidationError(
                message=_('Not an allowed ContentVariant for this user'),
                code='disallowed_contentvariant'
            )
        if self.strength > self.usercomponent.strength:
            raise ValidationError(
                _('Protection strength too low, required: %(strength)s'),
                code="strength",
                params={'strength': self.strength},
            )
        if (
            settings.DATABASES["default"]["ENGINE"] ==
            "django.db.backends.mysql"
        ):
            obj = AssignedContent.objects.filter(
                usercomponent=self.usercomponent,
                info=self.info
            ).first()

            if obj and obj.id != getattr(self, "id", None):
                raise ValidationError(
                    message=_("Unique Content already exists"),
                    code='unique_together',
                )
        super().clean()

    @property
    def is_hidden(self):
        return self.getflag("unlisted")



__all__ = (
    "UserTestMixin", "UCTestMixin", "EntityDeletionMixin", "DefinitionsMixin"
)
import logging
from urllib.parse import quote_plus

from datetime import timedelta

from django.contrib.auth.mixins import AccessMixin
from django.shortcuts import get_object_or_404
from django.contrib.auth import get_user_model, REDIRECT_FIELD_NAME
from django.http import HttpResponseRedirect, HttpResponseServerError
from django.http.response import HttpResponseBase
from django.utils import timezone
from django.conf import settings
from django.urls import reverse_lazy
from django.utils.translation import gettext
from django.forms.widgets import Media

from ..helpers import merge_get_url
from ..constants import (
    VariantType, TokenCreationError, ProtectionType,
    loggedin_active_tprotections
)
from ..models import UserComponent, AuthToken, TravelProtection

logger = logging.getLogger(__name__)


class DefinitionsMixin(object):

    def get_context_data(self, **kwargs):
        kwargs["raw_update_type"] = VariantType.raw_update.value
        kwargs["component_feature_type"] = VariantType.component_feature.value
        kwargs["content_feature_type"] = VariantType.content_feature.value
        kwargs["no_export_type"] = VariantType.no_export.value
        kwargs["hostpart"] = "{}://{}".format(
            self.request.scheme, self.request.get_host()
        )
        kwargs.setdefault("media", Media())
        return super().get_context_data(**kwargs)


class UserTestMixin(DefinitionsMixin, AccessMixin):
    preserved_GET_parameters = {"token", "search", "id", "protection"}
    login_url = reverse_lazy(getattr(
        settings,
        "LOGIN_URL",
        "auth:login"
    ))
    # don't allow AccessMixin to redirect, handle in test_token
    raise_exception = True
    _travel_request = None

    def dispatch_extra(self, request, *args, **kwargs):
        return None

    def dispatch(self, request, *args, **kwargs):
        _ = gettext
        self.request.is_owner = getattr(self.request, "is_owner", False)
        self.request.is_special_user = \
            getattr(self.request, "is_special_user", False)
        self.request.is_staff = getattr(self.request, "is_staff", False)
        self.request.auth_token = getattr(self.request, "auth_token", None)
        try:
            user_test_result = self.test_func()
        except TokenCreationError:
            logger.exception("Token creation failed")
            return HttpResponseServerError(
                _("Token creation failed, try again")
            )
        if isinstance(user_test_result, HttpResponseBase):
            return user_test_result
        elif not user_test_result:
            return self.handle_no_permission()

        ret = self.dispatch_extra(request, *args, **kwargs)
        if ret:
            return ret
        return super().dispatch(request, *args, **kwargs)

    def sanitize_GET(self):
        GET = self.request.GET.copy()
        for key in list(GET.keys()):
            if key not in self.preserved_GET_parameters:
                GET.pop(key, None)
        return GET

    def get_travel_for_request(self):
        if self._travel_request is None:
            self._travel_request = \
                TravelProtection.objects.get_active_for_request(self.request)
        return self._travel_request

    def get_context_data(self, **kwargs):
        kwargs["sanitized_GET"] = self.sanitize_GET()
        if kwargs["sanitized_GET"]:
            kwargs["sanitized_GET"] = "{}&".format(
                kwargs["sanitized_GET"].urlencode()
            )
        else:
            kwargs["sanitized_GET"] = ""
        return super().get_context_data(**kwargs)

    # by default only owner with login can access view
    def test_func(self):
        return self.has_special_access(
            user_by_login=True
        )

    def replace_token(self):
        GET = self.request.GET.copy()
        GET["token"] = self.request.auth_token.token
        return HttpResponseRedirect(
            redirect_to="?".join((self.request.path, GET.urlencode()))
        )

    def create_token(self, extra=None):
        d = {
            "usercomponent": self.usercomponent,
            "session_key": None,
            "extra": {}
        }
        if "token" not in self.request.GET:
            d["session_key"] = self.request.session.session_key

        token = AuthToken(**d)
        # to copy dictionary
        if extra:
            token.extra.update(extra)
        token.save()
        return token

    def remove_old_tokens(self, expire=None):
        if not expire:
            expire = timezone.now()-self.usercomponent.token_duration
        return self.usercomponent.authtokens.filter(
            created__lt=expire, persist=-1
        ).delete()

    def test_token(self, minstrength=0, force_token=False, taint=False):
        expire = timezone.now()-self.usercomponent.token_duration
        no_token = not force_token and self.usercomponent.required_passes == 0
        ptype = ProtectionType.access_control.value
        if minstrength >= 4:
            no_token = False
            ptype = ProtectionType.authentication.value

        # delete old token, so no confusion happen
        self.remove_old_tokens(expire)

        # only valid tokens here
        tokenstring = self.request.GET.get("token", None)
        token = None
        if tokenstring:
            # find by tokenstring
            token = self.usercomponent.authtokens.filter(
                token=tokenstring
            ).first()
        elif self.request.session.session_key:
            # use session_key
            token = self.usercomponent.authtokens.filter(
                session_key=self.request.session.session_key
            ).first()
        elif not no_token:
            # generate session key if it not exist and token is required
            self.request.session.cycle_key()
        if token and token.extra.get("prot_strength", 0) >= minstrength:
            self.request.token_expires = \
                token.created+self.usercomponent.token_duration
            # case will never enter
            # if not token.session_key and "token" not in self.request.GET:
            #     return self.replace_token()
            if (
                token.extra["prot_strength"] >= 4 and
                not token.extra.get("taint", False)
            ):
                self.request.is_special_user = True
                self.request.is_owner = True
            self.request.auth_token = token
            return True

        # if result is impossible and token invalid try to login
        if minstrength >= 4 and not self.usercomponent.can_auth:
            # auth won't work for logged in users
            if self.request.user.is_authenticated:
                return False
            # remove token and redirect
            # login_url can be also on a different host => merge_get_url
            target = "{}?{}={}".format(
                self.get_login_url(),
                REDIRECT_FIELD_NAME,
                quote_plus(
                    merge_get_url(
                        self.request.build_absolute_uri(),
                        token=None
                    )
                )
            )
            return HttpResponseRedirect(redirect_to=target)

        protection_codes = None
        if "protection" in self.request.GET:
            protection_codes = self.request.GET.getlist("protection")
        # execute protections for side effects even no_token
        self.request.protections = self.usercomponent.auth(
            request=self.request, scope=self.scope,
            protection_codes=protection_codes, ptype=ptype
        )
        if (
            type(self.request.protections) is int and  # because: False==0
            self.request.protections >= minstrength
        ):
            # generate only tokens if required
            if no_token:
                return True

            token = self.create_token(
                extra={
                    "strength": self.usercomponent.strength,
                    "prot_strength": self.request.protections,
                    "taint": taint
                }
            )

            if (
                token.extra["prot_strength"] >= 4 and
                not token.extra.get("taint", False)
            ):
                self.request.is_special_user = True
                self.request.is_owner = True

            self.request.token_expires = \
                token.created+self.usercomponent.token_duration
            self.request.auth_token = token
            if "token" in self.request.GET:
                return self.replace_token()
            return True
        return False

    def has_special_access(
        self, user_by_login=True, user_by_token=False,
        staff=False, superuser=False
    ) -> bool:
        if self.request.user.is_authenticated:
            t = self.get_travel_for_request().filter(
                login_protection__in=loggedin_active_tprotections
            ).exists()
            # auto activate but not deactivate
            if t:
                self.request.session["is_travel_protected"] = t
        if not hasattr(self, "usercomponent"):
            self.usercomponent = self.get_usercomponent()
        if user_by_login and self.request.user == self.usercomponent.user:
            self.request.is_owner = True
            self.request.is_special_user = True
            return True
        # can use auth token != X-TOKEN
        if user_by_token and self.test_token(4) is True:
            return True

        # remove user special state if is_travel_protected
        if self.request.session.get("is_travel_protected", False):
            return False
        if superuser and self.request.user.is_superuser:
            self.request.is_special_user = True
            self.request.is_staff = True
            return True
        if staff and self.request.user.is_staff:
            if type(staff) is bool:
                self.request.is_special_user = True
                self.request.is_staff = True
                return True
            if not isinstance(staff, (tuple, list, set)):
                staff = (staff,)

            if not all(
                map(self.request.user.has_perm, staff)
            ):
                return False
            self.request.is_special_user = True
            self.request.is_staff = True
            return True
        return False

    def get_user(self):
        """ Get user from user field or request """
        if (
                "user" not in self.kwargs and
                self.request.user.is_authenticated
           ):
            return self.request.user

        model = get_user_model()
        margs = {model.USERNAME_FIELD: None}
        margs[model.USERNAME_FIELD] = self.kwargs.get("user", None)
        return get_object_or_404(
            model.objects.select_related("spider_info"),
            **margs
        )

    def get_usercomponent(self):
        query = {
            "token": self.kwargs["token"]
        }
        return get_object_or_404(
            UserComponent.objects.prefetch_related(
                "authtokens", "protections"
            ),
            **query
        )

    def handle_no_permission(self):
        # in case no protections are used (e.g. add content)
        p = getattr(self.request, "protections", False)
        if not bool(p):
            # return 403
            return super().handle_no_permission()
        # should be never true here
        assert(p is not True)
        context = {
            "sanitized_GET": self.sanitize_GET(),
            "LOGIN_URL": self.get_login_url(),
            "scope": getattr(self, "scope", None),
            "uc": self.usercomponent,
            "object": getattr(self, "object", None),
            "is_public_view": self.usercomponent.public
        }
        assert len(p) > 0
        return self.response_class(
            request=self.request,
            template=self.get_noperm_template_names(),
            # render with own context; get_context_data may breaks stuff or
            # disclose informations
            context=context,
            using=self.template_engine,
            content_type=self.content_type
        )

    def get_noperm_template_names(self):
        return "spider_base/protections/protections.html"


class UCTestMixin(UserTestMixin):
    usercomponent = None

    def dispatch(self, request, *args, **kwargs):
        self.usercomponent = self.get_usercomponent()
        return super(UCTestMixin, self).dispatch(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        # for protections & contents
        return self.get(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        # for protections & contents
        return self.get(request, *args, **kwargs)


class EntityDeletionMixin(UserTestMixin):
    object = None
    http_method_names = ['get', 'post', 'delete']

    def get_context_data(self, **kwargs):
        _time = self.get_required_timedelta()
        if _time and self.object.deletion_requested:
            now = timezone.now()
            if self.object.deletion_requested + _time >= now:
                kwargs["remaining"] = timedelta(seconds=0)
            else:
                kwargs["remaining"] = self.object.deletion_requested+_time-now
        return super().get_context_data(**kwargs)

    def get_required_timedelta(self):
        _time = self.object.deletion_period
        if _time:
            _time = timedelta(seconds=_time)
        else:
            _time = timedelta(seconds=0)
        return _time

    def options(self, request, *args, **kwargs):
        ret = super().options()
        ret["Access-Control-Allow-Origin"] = self.request.get_host()
        ret["Access-Control-Allow-Methods"] = "POST, GET, DELETE, OPTIONS"
        return ret

    def delete(self, request, *args, **kwargs):
        _time = self.get_required_timedelta()
        if _time:
            now = timezone.now()
            if self.object.deletion_requested:
                if self.object.deletion_requested+_time >= now:
                    return self.get(request, *args, **kwargs)
            else:
                self.object.deletion_requested = now
                self.object.save()
                return self.get(request, *args, **kwargs)
        self.object.delete()
        return HttpResponseRedirect(self.get_success_url())

    def post(self, request, *args, **kwargs):
        # delete works if allowed in CORS
        if request.POST.get("action") == "reset":
            return self.reset(request, *args, **kwargs)
        elif request.POST.get("action") == "delete":
            return self.delete(request, *args, **kwargs)
        return super().get(request, *args, **kwargs)

    def reset(self, request, *args, **kwargs):
        self.object.deletion_requested = None
        self.object.save(update_fields=["deletion_requested"])
        return HttpResponseRedirect(self.get_success_url())



__all__ = (
    "SignupView", "LoginView", "LogoutView", "UpdateView",
    "PasswordChangeView", "PasswordChangeDoneView", "SignupDoneView"
)

from django.http import HttpResponseRedirect
from django.contrib.auth import login, authenticate
from django.contrib.auth.views import (
    LoginView as _LoginView,
    LogoutView as _LogoutView,
    PasswordChangeView as _PasswordChangeView,
    PasswordChangeDoneView as _PasswordChangeDoneView
)
from django.views.generic.edit import FormView, UpdateView
from django.views.generic.base import TemplateView
from django.contrib.auth.mixins import LoginRequiredMixin
from django.views.decorators.cache import never_cache
from django.utils.decorators import method_decorator
from django.urls import reverse_lazy
from django.contrib.auth import get_user_model
from django.conf import settings

from .forms import SignupForm, UserUpdateForm
from spkcspider.apps.spider.forms import SpiderAuthForm
from spkcspider.apps.spider.views import DefinitionsMixin

# Create your views here.


class PasswordChangeView(DefinitionsMixin, _PasswordChangeView):
    pass


class PasswordChangeDoneView(DefinitionsMixin, _PasswordChangeDoneView):
    pass


class SignupView(DefinitionsMixin, FormView):
    template_name = 'registration/signup.html'
    form_class = SignupForm
    success_url = reverse_lazy("auth:signup_done")

    def form_valid(self, form):
        form.save()
        username = form.cleaned_data.get('username')
        raw_password = form.cleaned_data.get('password1')
        user = authenticate(username=username, password=raw_password)
        login(self.request, user)
        return HttpResponseRedirect(self.get_success_url())

    def get_success_url(self):
        if "next" in self.request.GET:
            return self.request.GET["next"]
        return self.success_url


class SignupDoneView(DefinitionsMixin, TemplateView):
    template_name = 'registration/thanks.html'


class LoginView(DefinitionsMixin, _LoginView):
    form_class = SpiderAuthForm

    def get(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            return HttpResponseRedirect(self.get_success_url())
        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return super().post(request, *args, **kwargs)

    def form_valid(self, form):
        """Security check complete. Log the user in."""
        # backend for loging in, select the first
        backend = settings.AUTHENTICATION_BACKENDS[0]
        login(
            self.request,
            form.get_user(),
            backend
        )
        return HttpResponseRedirect(self.get_success_url())

    def form_invalid(self, form):
        form.reset_protections()
        return super().form_invalid(form)


class LogoutView(DefinitionsMixin, _LogoutView):
    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        ret = super().dispatch(request, *args, **kwargs)
        ret["Clear-Site-Data"] = "*"
        return ret


class UpdateView(DefinitionsMixin, LoginRequiredMixin, UpdateView):
    success_url = reverse_lazy('auth:profile')
    template_name = 'registration/profile.html'
    form_class = UserUpdateForm

    def get_object(self, queryset=None):
        # clone user object
        return get_user_model().objects.get(pk=self.request.user.pk)



__all__ = (
    "AdminTokenManagement", "TokenDeletionRequest", "TokenRenewal",
    "ConfirmTokenUpdate", "RequestTokenUpdate"
)

import logging

from django.conf import settings
from django.shortcuts import get_object_or_404
from django.db.models import Q
from django.views.generic.base import View, TemplateView
from django.views.generic.edit import DeleteView
from django.utils.translation import gettext

from django.http import (
    Http404, HttpResponseServerError, JsonResponse, HttpResponseRedirect,
    HttpResponse
)

from django.utils.http import is_safe_url
from django.views.decorators.csrf import csrf_exempt
from django.utils.decorators import method_decorator
from django.contrib import messages
from django.test import Client


import requests

from ._core import UCTestMixin, UserTestMixin
from ._referrer import ReferrerMixin
from ..models import AuthToken
from ..helpers import get_settings_func, get_requests_params, merge_get_url
from ..constants import TokenCreationError


logger = logging.getLogger(__name__)


class AdminTokenManagement(UCTestMixin, View):
    scope = None
    created_token = None

    def dispatch_extra(self, request, *args, **kwargs):
        self.remove_old_tokens()

    def test_func(self):
        if self.scope == "delete":
            return self.has_special_access(
                user_by_login=True
            )
        else:
            return self.has_special_access(
                user_by_login=True, user_by_token=True
            )

    def post(self, request, *args, **kwargs):
        if self.request.POST.get("add_token"):
            extra = {
                "strength": self.usercomponent.strength,
            }
            if self.request.POST.get("restrict"):
                extra["ids"] = [-1]
            elif self.request.GET.get("id", "") != "":
                try:
                    extra["ids"] = list(set(map(
                        int, self.request.GET.getlist("id")
                    )))
                except Exception:
                    return HttpResponse(status=400)
            if self.request.GET.get("search", "") != "":
                extra["search"] = list(set(
                    self.request.GET.getlist("search")
                ))
            self.created_token = self.create_token(
                extra=extra
            )

        try:
            delids = set(map(int, self.request.POST.getlist("delete_tokens")))
        except Exception:
            return HttpResponse(status=400)
        if delids:
            delquery = AuthToken.objects.filter(
                usercomponent=self.usercomponent,
                id__in=delids
            )
            if self.scope == "share":
                delquery = delquery.filter(
                    Q(session_key=request.session.session_key) |
                    Q(token=request.auth_token)
                )
            shall_redirect = delquery.filter(token=request.auth_token).exists()
            delquery.delete()
            del delquery
            if shall_redirect:
                return HttpResponseRedirect(
                    redirect_to=request.get_full_path()
                )

        return self.get(request, *args, **kwargs)

    def _token_dict(self, token):
        # sl or not initialized map to full
        full = "sl" in token.extra.get("intentions", ["sl"])
        ret = {
            "admin_key": token.extra.get("prot_strength", 0) >= 4,
            "created": self.created_token == token,
            "restricted": token.extra.get("ids", []) == [-1],
            "expires": None if token.persist >= 0 else (
                token.created +
                self.usercomponent.token_duration
            ).strftime("%a, %d %b %Y %H:%M:%S %z"),
            "referrer": token.referrer.url if token.referrer else "",
            "token": token.token if full else None,
            "name": str(token),
            "id": token.id,
            "needs_confirmation": (
                token.extra.get("request_intentions") is not None or
                token.extra.get("request_search") is not None or
                token.extra.get("request_referrer") is not None
            )
        }
        if token.session_key == self.request.session.session_key:
            assert(not token.referrer)
            ret["same_session"] = True
        elif self.request.auth_token == token:
            assert(not token.referrer)
            ret["same_session"] = True
        else:
            ret["same_session"] = False

        return ret

    def get(self, request, *args, **kwargs):
        # don't filter here, as used twice
        if self.scope == "delete":
            response = {
                "tokens": list(map(
                    self._token_dict,
                    AuthToken.objects.filter(
                        usercomponent=self.usercomponent
                    )
                )),
            }
        else:
            response = {
                "tokens": list(map(
                    self._token_dict,
                    AuthToken.objects.filter(
                        Q(session_key=request.session.session_key) |
                        Q(token=request.auth_token),
                        usercomponent=self.usercomponent,
                    )
                )),
            }
        return JsonResponse(response)


class TokenDeletionRequest(UCTestMixin, DeleteView):
    model = AuthToken
    template_name = "spider_base/protections/authtoken_confirm_delete.html"

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        try:
            return super().dispatch(request, *args, **kwargs)
        except Http404:
            return get_settings_func(
                "SPIDER_RATELIMIT_FUNC",
                "spkcspider.apps.spider.functions.rate_limit_default"
            )(self, request)

    def test_func(self):
        # tokens which are not persistent can be easily deleted
        if self.object.persist < 0:
            return True
        if self.has_special_access(
            user_by_token=True, user_by_login=True,
            superuser=True, staff="spider_base.delete_authtoken"
        ):
            return True
        return False

    def get_usercomponent(self):
        return self.object.usercomponent

    def get_object(self, queryset=None):
        if not queryset:
            queryset = self.get_queryset()

        return get_object_or_404(
            queryset,
            token=self.request.GET.get("delete", ""),
        )

    def get(self, request, *args, **kwargs):
        # tokens which are not persistent can be easily deleted
        if self.object.persist < 0:
            return self.delete(request, *args, **kwargs)
        return super().get(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        url = None
        if self.object.referrer:
            url = self.object.referrer.url
        self.object.delete()
        if not url:
            return HttpResponse(status=200)
        return HttpResponseRedirect(
            redirect_to=url
        )

    def post(self, request, *args, **kwargs):
        return self.delete(request, *args, **kwargs)

    def options(self, request, *args, **kwargs):
        ret = super().options()
        ret["Access-Control-Allow-Origin"] = "*"
        ret["Access-Control-Allow-Methods"] = "GET, POST, DELETE, OPTIONS"
        return ret


class TokenRenewal(UCTestMixin, View):
    model = AuthToken
    oldtoken = None

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        try:
            return super().dispatch(request, *args, **kwargs)
        except Http404:
            return get_settings_func(
                "SPIDER_RATELIMIT_FUNC",
                "spkcspider.apps.spider.functions.rate_limit_default"
            )(self, request)

    def get_usercomponent(self):
        token = self.request.POST.get("token", None)
        if not token:
            raise Http404()
        self.request.auth_token = get_object_or_404(
            AuthToken,
            token=token,
            persist__gte=0,
            referrer__isnull=False
        )
        assert (
            "persist" in self.request.auth_token.extra.get(
                "intentions", []
            )
        )
        return self.request.auth_token.usercomponent

    def test_func(self):
        return True

    def options(self, request, *args, **kwargs):
        ret = super().options()
        ret["Access-Control-Allow-Origin"] = \
            self.request.auth_token.referrer.host
        ret["Access-Control-Allow-Methods"] = "POST, GET, OPTIONS"
        return ret

    def update_with_post(self):
        # application/x-www-form-urlencoded is best here,
        # for beeing compatible to most webservers
        # client side rdf is no problem
        # NOTE: csrf must be disabled or use csrf token from GET,
        #       here is no way to know the token value
        d = {
            "oldtoken": self.oldtoken,
            "token": self.request.auth_token.token,
            "hash_algorithm": settings.SPIDER_HASH_ALGORITHM.name,
            "action": "renew"
        }
        if "payload" in self.request.POST:
            d["payload"] = self.request.POST["payload"]

        params, can_inline = get_requests_params(
            self.request.auth_token.referrer.url
        )
        if can_inline:
            response = Client().post(
                self.request.auth_token.referrer.url,
                data=d,
                Connection="close",
                Referer="%s://%s" % (
                    self.request.scheme,
                    self.request.path
                )
            )
            if response.status_code >= 400:
                return False
        else:
            try:
                with requests.post(
                    self.request.auth_token.referrer.url,
                    data=d,
                    headers={
                        "Referer": "%s://%s" % (
                            self.request.scheme,
                            self.request.path
                        ),
                        "Connection": "close"
                    },
                    **params
                ) as resp:
                    resp.raise_for_status()
            except requests.exceptions.SSLError as exc:
                logger.info(
                    "referrer: \"%s\" has a broken ssl configuration",
                    self.request.auth_token.referrer, exc_info=exc
                )
                return False
            except Exception as exc:
                logger.info(
                    "post failed: \"%s\" failed",
                    self.request.auth_token.referrer, exc_info=exc
                )
                return False
        return True

    def post(self, request, *args, **kwargs):
        self.oldtoken = self.request.auth_token.token
        self.request.auth_token.initialize_token()
        success = True
        sl = "sl" in self.request.auth_token.extra.get("intentions", [])
        if sl:
            # only the original referer can access this
            success = (
                self.request.headers.get("Referer", "").startswith(
                    self.request.auth_token.referrer.url
                )
            )
        else:
            # if not serverless:
            # you cannot steal a token because it is bound to a domain
            # damage can be only done in persisted data
            success = self.update_with_post()
        if success:
            try:
                self.request.auth_token.save()
            except TokenCreationError:
                success = False
        if not success:
            logger.exception("Token creation failed")
            return HttpResponseServerError(
                "Token update failed, try again"
            )
        if sl:
            ret = HttpResponse(
                self.request.auth_token.token.encode(
                    "ascii"
                ),
                content_type="text/plain",
                status=200
            )
        else:
            ret = HttpResponse(status=200)
        # no sl: simplifies update logic for thirdparty web servers
        # sl: safety (but anyway checked by referer check)
        ret["Access-Control-Allow-Origin"] = \
            self.request.auth_token.referrer.host
        return ret


class ConfirmTokenUpdate(ReferrerMixin, UCTestMixin, TemplateView):
    model = AuthToken
    redirect_field_name = "next"
    preserved_GET_parameters = {"token", "page", "search", "id", "protection"}

    def get_redirect_url(self, sanitized_GET=None):
        """Return the user-originating redirect URL if it's safe."""
        if not sanitized_GET:
            sanitized_GET = self.sanitize_GET()
        redirect_to = self.request.GET.get(
            self.redirect_field_name,
            self.request.GET.get(self.redirect_field_name, '')
        )
        url_is_safe = is_safe_url(
            url=redirect_to,
            allowed_hosts={self.request.get_host()},
            require_https=self.request.is_secure(),
        )
        redirect_to = "{}?{}".format(
            url_is_safe.rstrip("?&"),
            sanitized_GET
        )

        return redirect_to if url_is_safe else ''

    def test_func(self):
        return self.has_special_access(
            user_by_login=True, user_by_token=True
        )

    def dispatch_extra(self, request, *args, **kwargs):
        _ = gettext
        context = self.get_context_data()
        # fallback to intentions if no request_intentions
        # remove "domain", "sl"
        context["intentions"] = set(self.object.extra.get(
            "request_intentions", self.object.extra.get(
                "intentions", []
            )
        ))
        # sl or domain should NOT be allowed for token updates
        # sl is only for clients, domain only initial and should be removed
        context["intentions"].difference_update({"domain", "sl"})
        context["action"] = "update"
        context["uc"] = self.object.usercomponent

        rreferrer = request.POST.get("referrer", None)
        if rreferrer:
            context["referrer"] = merge_get_url(rreferrer)
            if not get_settings_func(
                "SPIDER_URL_VALIDATOR",
                "spkcspider.apps.spider.functions.validate_url_default"
            )(context["referrer"], self):
                context["action"] = "referrer_invalid"
            # for donotact
            if self.object.referrer and self.object.referrer.url == rreferrer:
                rreferrer = None
        else:
            rreferrer = self.object.extra.get("request_referrer", None)
            if rreferrer:
                context["referrer"] = merge_get_url(rreferrer)
                if not get_settings_func(
                    "SPIDER_URL_VALIDATOR",
                    "spkcspider.apps.spider.functions.validate_url_default"
                )(context["referrer"], self):
                    return HttpResponse(
                        status=400,
                        content=_('Insecure url: %(url)s') % {
                            "url": context["referrer"]
                        }
                    )
            elif self.object.referrer:
                context["referrer"] = self.object.referrer.url
            else:
                context["referrer"] = ""
        context["ids"] = self.object.usercomponent.contents.values_list(
            "id", flat=True
        )
        context["ids"] = set(context["ids"])
        # if requested referrer is available DO delete invalid and DO care
        ret = self.handle_referrer_request(
            context, self.object, dontact=not rreferrer, no_oldtoken=True
        )
        if isinstance(ret, HttpResponseRedirect):
            if context.get("post_success", False):
                messages.success(request, _("Intention update successful"))
            else:
                messages.error(request, _("Intention update failed"))
            return HttpResponseRedirect(self.get_redirect_url(
                context["sanitized_GET"]
            ))
        return ret

    def get_usercomponent(self):
        self.object = get_object_or_404(
            AuthToken,
            id=self.request.GET.get("refid", None),
        )
        return self.object.usercomponent


class RequestTokenUpdate(UserTestMixin, View):

    def test_func(self):
        self.request.auth_token = get_object_or_404(
            AuthToken,
            token=self.request.POST("token"),
        )
        return bool(self.request.auth_token)

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        try:
            return super().dispatch(request, *args, **kwargs)
        except Http404:
            return get_settings_func(
                "SPIDER_RATELIMIT_FUNC",
                "spkcspider.apps.spider.functions.rate_limit_default"
            )(self, request)

    def get(self, request, *args, **kwargs):
        ret_dict = {
            "intentions": request.auth_token.extra.get("intentions") or [],
            "search": request.auth_token.extra.get("search") or [],
            "referrer": None
        }
        if request.auth_token.referrer:
            ret_dict["referrer"] = request.auth_token.referrer.url
        if request.auth_token.extra.get("request_referrer") is not None:
            ret_dict["referrer"] = \
                request.auth_token.extra["request_referrer"]
        if request.auth_token.extra.get("request_intentions") is not None:
            ret_dict["intentions"] = \
                request.auth_token.extra["request_intentions"]
        if request.auth_token.extra.get("request_search") is not None:
            ret_dict["search"] = request.auth_token.extra["request_search"]
        return JsonResponse(ret_dict)

    def post(self, request, *args, **kwargs):
        if "referrer" in request.POST:
            # will be checked a second time
            request.auth_token.extra["request_referrer"] = \
                request.POST.get("referrer")
        elif not request.auth_token.referrer:
            return HttpResponse("no referrer", status=400)
        if request.POST.get("intentions", "") != "":
            # will be checked a second time
            request.auth_token.extra["request_intentions"] = \
                request.POST.getlist("intentions")
        if request.POST.get("search", "") != "":
            # will be checked a second time
            request.auth_token.extra["request_search"] = \
                request.POST.getlist("search")
        request.auth_token.save(update_fields=["extra"])
        return self.get(request, *args, **kwargs)

    def options(self, request, *args, **kwargs):
        ret = super().options()
        ret["Access-Control-Allow-Methods"] = "POST, GET, OPTIONS"
        return ret



# Generated by Django 2.2.1 on 2019-05-30 21:49

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):


    dependencies = [
        ('spider_verifier', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='dataverificationtag',
            options={'permissions': [('can_verify', 'Can verify Data Tag?')]},
        ),
        migrations.AlterField(
            model_name='dataverificationtag',
            name='note',
            field=models.TextField(blank=True, default=''),
        ),
        migrations.AlterField(
            model_name='dataverificationtag',
            name='verification_state',
            field=models.CharField(choices=[('retrieve', 'retrieval pending'), ('pending', 'pending'), ('verified', 'verified'), ('invalid', 'invalid'), ('rejected', 'rejected')], default='retrieve', max_length=10),
        ),
        migrations.AddField(
            model_name='dataverificationtag',
            name='source',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='spider_verifier.VerifySourceObject'),
        ),
        migrations.AlterField(
            model_name='dataverificationtag',
            name='verification_state',
            field=models.CharField(choices=[('pending', 'pending'), ('verified', 'verified'), ('invalid', 'invalid'), ('rejected', 'rejected')], default='pending', max_length=10),
        ),
        migrations.AddField(
            model_name='verifysourceobject',
            name='token',
            field=models.CharField(blank=True, max_length=126, null=True, unique=True),
        ),
    ]



__all__ = ["KeyForm", "AnchorServerForm", "AnchorKeyForm"]
# "AnchorGovForm"

import binascii

from django.db import models
from django import forms
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext

from cryptography import exceptions
from cryptography.hazmat.primitives.asymmetric import padding

from spkcspider.apps.spider.fields import MultipleOpenChoiceField
from spkcspider.apps.spider.widgets import ListWidget
from spkcspider.apps.spider.models import TravelProtection, AssignedContent
from spkcspider.apps.spider.constants import loggedin_active_tprotections


from .models import PublicKey, AnchorServer, AnchorKey


_help_text_key = _(
    '"Public Key"-Content for signing identifier. It is recommended to use different keys for signing and encryption. '  # noqa: E501
    "Reason herefor is, that with a change of the signing key the whole anchor gets invalid and the signing key should be really carefully saved away. "  # noqa: E501
    "In contrast the encryption keys can be easily exchanged and should be available for encryption"  # noqa: E501
)


class KeyForm(forms.ModelForm):
    hash_algorithm = forms.CharField(
        widget=forms.HiddenInput(), disabled=True
    )
    setattr(hash_algorithm, "hashable", False)

    class Meta:
        model = PublicKey
        fields = ['key']

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.initial["hash_algorithm"] = settings.SPIDER_HASH_ALGORITHM.name
        setattr(self.fields['key'], "hashable", True)

    def clean_key(self):
        _ = gettext
        data = self.cleaned_data['key'].strip()
        if data == "":
            raise forms.ValidationError(
                _('Empty Key'),
                code="empty"
            )
        if "PRIVATE" in data.upper():
            raise forms.ValidationError(
                _('Private Key')
            )
        return data


class AnchorServerForm(forms.ModelForm):
    identifier = forms.CharField(disabled=True)
    anchor_type = forms.CharField(disabled=True, initial="url")
    setattr(identifier, "hashable", True)
    scope = ""
    old_urls = MultipleOpenChoiceField(
        widget=ListWidget(
            format_type="url", item_label=_("Url to superseded anchor")
        ), required=False
    )

    class Meta:
        model = AnchorServer
        fields = ["new_url", "old_urls"]

    def __init__(self, scope, **kwargs):
        self.scope = scope
        super().__init__(**kwargs)
        if self.scope == "add":
            del self.fields["identifier"]
            del self.fields["anchor_type"]
            del self.fields["new_url"]

    def clean_old_urls(self):
        values = self.cleaned_data["old_urls"]
        if not isinstance(values, list):
            raise forms.ValidationError(
                _("Invalid format"),
                code='invalid_format'
            )
        return values

    def clean(self):
        _ = gettext
        ret = super().clean()
        if ret.get("new_url", None) and ret.get("old_urls", []):
            raise forms.ValidationError(
                _(
                    "Specify either replacement url or "
                    "urls to superseding anchors"
                ),
                code='invalid_choices'
            )
        return ret


class AnchorKeyForm(forms.ModelForm):
    identifier = forms.CharField(disabled=True)
    anchor_type = forms.CharField(disabled=True, initial="signature")
    key = forms.ModelChoiceField(
        queryset=AssignedContent.objects.filter(
            ctype__name="PublicKey",
            info__contains="\x1epubkeyhash="
        )
    )

    scope = ""

    class Meta:
        model = AnchorKey
        fields = ['key', 'signature']

    field_order = ['identifier', 'signature', 'key']

    def __init__(self, scope, request, **kwargs):
        self.scope = scope
        super().__init__(**kwargs)
        if self.instance.id:
            self.initial["key"] = \
                self.instance.associated.attached_to_content
        if self.scope == "add":
            del self.fields["identifier"]
            del self.fields["signature"]
            del self.fields["anchor_type"]
            # self.fields["signature"].disabled = True
            # self.fields["signature"].required = False

        if self.scope in ("add", "update"):
            travel = TravelProtection.objects.get_active_for_request(request)
            travel = travel.filter(
                login_protection__in=loggedin_active_tprotections
            )
            self.fields["key"].queryset = \
                self.fields["key"].queryset.exclude(
                    models.Q(usercomponent__travel_protected__in=travel) |
                    models.Q(travel_protected__in=travel)
                ).filter(
                    usercomponent=self.instance.associated.usercomponent
                )
        elif self.scope in ("raw", "list", "view"):
            self.fields["key"] = forms.CharField(
                initial=self.instance.key.key,
                widget=forms.TextArea
            )
            setattr(self.fields['key'], "hashable", True)

    def clean(self):
        _ = gettext
        ret = super().clean()
        pubkey = self.cleaned_data.get("key")
        if pubkey:
            pubkey = pubkey.content.get_key_ob()
            if not pubkey:
                self.add_error("key", forms.ValidationError(
                    _("key not usable for signing"),
                    code="unusable_key"
                ))
        if self.scope == "add":
            ret["signature"] = "<replaceme>"
            return ret
        if not pubkey:
            return
        chosen_hash = settings.SPIDER_HASH_ALGORITHM
        try:
            pubkey.verify(
                binascii.unhexlify(ret["signature"]),
                ret["identifier"].encode("utf-8"),
                padding.PSS(
                    mgf=padding.MGF1(chosen_hash),
                    salt_length=padding.PSS.MAX_LENGTH
                ),
                chosen_hash
            )
        except exceptions.InvalidSignature:
            self.add_error("signature", forms.ValidationError(
                _("signature incorrect"),
                code="incorrect_signature"
            ))
        except (binascii.Error, KeyError, ValueError):
            self.add_error("signature", forms.ValidationError(
                _("signature malformed or missing"),
                code="malformed_signature"
            ))
        return ret

    def save(self, commit=True):
        self.instance.associated.attached_to_content = \
            self.cleaned_data["key"]
        return super().save(commit)


# class AnchorGovForm(forms.ModelForm):
#    identifier = forms.CharField(disabled=True, widget=forms.HiddenInput())
#    anchor_type = forms.CharField(disabled=True, initial="gov")
#    scope = ""
#
#    class Meta:
#        model = AnchorGov
#        fields = ['idtype', 'token']
#
#    def __init__(self, scope, **kwargs):
#        self.scope = scope
#        super().__init__(**kwargs)
#
#    def clean(self):
#        self.cleaned_data["type"] = "gov"
#        self.cleaned_data["identifier"] = self.instance.get_identifier()
#        self.cleaned_data["format"] = "gov_{verifier}_{token}"
#        if self.scope not in ["add", "update"]:
#            self.cleaned_data["verifier"] = ID_VERIFIERS[self.idtype]
#        del self.cleaned_data["idtype"]
#        return self.cleaned_data



# Generated by Django 2.1 on 2018-08-15 12:05

from django.db import migrations, models
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('spider_tags', '0002_auto_20180815_1128'),
    ]

    operations = [
        migrations.AlterField(
            model_name='spidertag',
            name='primary',
            field=models.BooleanField(blank=True, default=False),
        ),
        migrations.AlterField(
            model_name='spidertag',
            name='tagdata',
            field=jsonfield.fields.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name='spidertag',
            name='verified_by',
            field=jsonfield.fields.JSONField(blank=True, default=list),
        ),
    ]




from django.conf import settings
from django.urls import path, reverse_lazy


from django.views.generic.base import RedirectView

from .views import (
    SignupView, UpdateView, LoginView, LogoutView,
    PasswordChangeView, PasswordChangeDoneView, SignupDoneView
)


app_name = "auth"

# no recovery, only authentication

urlpatterns = [
    path('login/', LoginView.as_view(), name='login'),
    path('logout/', LogoutView.as_view(), name='logout'),
    path(
        'password_change/',
        PasswordChangeView.as_view(
            success_url=reverse_lazy("auth:password_change_done")
        ), name='password_change'
    ),
    path(
        'password_change/done/',
        PasswordChangeDoneView.as_view(), name='password_change_done'
    ),
    path('profile/', UpdateView.as_view(), name="profile"),
    path(
        '',
        RedirectView.as_view(
            url=reverse_lazy('auth:profile'),
            permanent=True
        )
    ),
]
if getattr(settings, "OPEN_FOR_REGISTRATION", False):
    urlpatterns.append(
        path('signup/', SignupView.as_view(), name="signup")
    )
    urlpatterns.append(
        path(
            'thanks/',
            SignupDoneView.as_view(), name="signup_done"
        )
    )



# Generated by Django 2.1.7 on 2019-02-26 18:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0002_20190518_squashed'),
        ('spider_tags', '0005_spidertag_updateable_by'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='spidertag',
            name='updateable_by',
        ),
        migrations.AddField(
            model_name='spidertag',
            name='updateable_by',
            field=models.ManyToManyField(blank=True, related_name='tags', to='spider_base.ReferrerObject'),
        ),
    ]



# Generated by Django 2.1.5 on 2019-02-11 20:15

from django.db import migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('spider_tags', '0004_auto_20190117_1253'),
    ]

    operations = [
        migrations.AddField(
            model_name='spidertag',
            name='updateable_by',
            field=jsonfield.fields.JSONField(blank=True, default=list),
        ),
    ]



# Generated by Django 2.2.2 on 2019-06-21 19:54

from django.db import migrations


def move_links(apps, schema_editor):
    AssignedContent = apps.get_model("spider_base", "AssignedContent")
    LinkContent = apps.get_model("spider_base", "LinkContent")

    for i in AssignedContent.objects.filter(
        ctype__name="Link"
    ):
        content = LinkContent.objects.get(id=i.object_id)
        i.attached_to_content = content.content
        i.save(update_fields=["attached_to_content"])


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0002_20190518_squashed'),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.RunPython(move_links),
    ]



# Generated by Django 2.2.3 on 2019-07-08 13:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_webcfg', '0002_auto_20190120_1252'),
    ]

    operations = [
        migrations.AlterField(
            model_name='webconfig',
            name='config',
            field=models.BinaryField(blank=True, default=b'', editable=True),
        ),
    ]



""" Content Views """

__all__ = (
    "ContentIndex", "ContentAdd", "ContentAccess", "ContentDelete",
    "TravelProtectionManagement"
)

from datetime import timedelta
from html import escape

from django.db.models.deletion import ProtectedError
from django.views.generic.edit import DeleteView, UpdateView, CreateView
from django.views.generic.list import ListView
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.http.response import HttpResponseBase
from django.http import Http404, HttpResponse, HttpResponseRedirect
from django.db import models
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.contrib import messages
from django.utils.translation import gettext
from django.views.decorators.csrf import csrf_exempt
from django.utils.decorators import method_decorator
from django.utils.safestring import mark_safe
from django.forms.widgets import Media

from next_prev import next_in_order, prev_in_order

from rdflib import Graph, Literal, URIRef, XSD


from ._core import UCTestMixin, EntityDeletionMixin, UserTestMixin
from ._referrer import ReferrerMixin

from ..models import (
    AssignedContent, ContentVariant, UserComponent, TravelProtection
)
from ..forms import UserContentForm, TravelProtectionManagementForm
from ..helpers import get_settings_func, add_property, merge_get_url
from ..queryfilters import (
    filter_contents, listed_variants_q, machine_variants_q
)
from ..constants import (
    spkcgraph, VariantType, static_token_matcher, loggedin_active_tprotections
)
from ..serializing import paginate_stream, serialize_stream

_forbidden_scopes = frozenset(["add", "list", "raw", "delete", "anchor"])

_extra = '' if settings.DEBUG else '.min'


class ContentBase(UCTestMixin):
    model = AssignedContent
    scope = None
    object = None
    # use token of content object instead
    no_token_usercomponent = True

    def dispatch(self, request, *args, **kwargs):
        try:
            return super().dispatch(request, *args, **kwargs)
        except Http404:
            return get_settings_func(
                "SPIDER_RATELIMIT_FUNC",
                "spkcspider.apps.spider.functions.rate_limit_default"
            )(self, request)

    def get_template_names(self):
        if self.scope in ("add", "update"):
            return ['spider_base/assignedcontent_form.html']
        elif self.scope == "list":
            return ['spider_base/assignedcontent_list.html']
        else:
            return ['spider_base/assignedcontent_access.html']

    def get_ordering(self, issearching=False):
        if self.scope != "list":
            # export: also serializer, other scopes: only one object, overhead
            return None
        # ordering will happen in serializer
        if "raw" in self.request.GET:
            return None
        return ("-priority", "-modified")

    def get_context_data(self, **kwargs):
        kwargs["request"] = self.request
        kwargs["scope"] = self.scope
        kwargs["uc"] = self.usercomponent
        kwargs["media"] = Media(
            css={
                "all": [
                    'node_modules/selectize/dist/css/selectize.default.css'
                ]
            },
            js=[
                'node_modules/qrcode-generator/qrcode.js',
                'node_modules/jquery/dist/jquery%s.js' % _extra,
                'node_modules/selectize/dist/js/standalone/selectize%s.js' % _extra  # noqa: E501
            ]
        )
        kwargs["enctype"] = "multipart/form-data"
        return super().get_context_data(**kwargs)

    def get_queryset(self):
        ret = self.model.objects.all()
        # skip search if user and single object
        if self.scope in ("add", "update", "raw_update"):
            return ret

        if getattr(self.request, "auth_token", None):
            idlist = \
                set(self.request.auth_token.extra.get("ids") or [])
            searchlist = \
                self.request.auth_token.extra.get("filter") or []
        else:
            idlist = set()
            searchlist = []

        if self.scope == "list":
            if "search" in self.request.POST or "id" in self.request.POST:
                searchlist += self.request.POST.getlist("search")
                if idlist:
                    idlist.intersection(self.request.POST.getlist("id"))
                else:
                    idlist.update(self.request.POST.getlist("id"))
            else:
                searchlist += self.request.GET.getlist("search")
                if idlist:
                    idlist.intersection(self.request.POST.getlist("id"))
                else:
                    idlist.update(self.request.POST.getlist("id"))
        elif self.scope not in ("add", "update", "raw_update"):
            searchlist += self.request.GET.getlist("search")

        # list only unlisted if explicity requested or export or:
        # if it has high priority (only for special users)
        # listing prioritized, unlisted content is different to the broader
        # search
        filter_unlisted = False
        if self.request.is_special_user:
            # all other scopes than list can show here _unlisted
            # this includes export
            if self.scope == "list" and "_unlisted" not in searchlist:
                filter_unlisted = 0
        else:
            filter_unlisted = True

        filter_q, counter = filter_contents(
            searchlist, idlist, filter_unlisted
        )
        if self.scope == "export":
            ret = ret.exclude(
                ctype__ctype__contains=VariantType.no_export.value
            )

        order = self.get_ordering(counter > 0)
        # distinct required?
        ret = ret.filter(filter_q)
        if order:
            ret = ret.order_by(*order)
        return ret


class ContentIndex(ReferrerMixin, ContentBase, ListView):
    model = AssignedContent
    scope = "list"
    no_token_usercomponent = False
    rate_limit_group = "spider_static_token_error"
    preserved_GET_parameters = {"token", "search", "id"}

    def dispatch_extra(self, request, *args, **kwargs):
        self.allow_domain_mode = ContentVariant.objects.filter(
            feature_for_components=self.usercomponent,
            name="DomainMode"
        ).exists()
        if "referrer" in self.request.GET:
            self.object_list = self.get_queryset()
            return self.handle_referrer()
        return None

    def get_queryset(self):
        travel = self.get_travel_for_request()
        t_ids = travel.values_list("associated_rel__id", flat=True)
        travel = travel.filter(
            login_protection__in=loggedin_active_tprotections
        )
        return super().get_queryset().filter(
            usercomponent=self.usercomponent
        ).exclude(
            models.Q(travel_protected__in=travel) |
            models.Q(id__in=t_ids)
        )

    def get_usercomponent(self):
        travel = self.get_travel_for_request().filter(
            login_protection__in=loggedin_active_tprotections
        )
        q = UserComponent.objects.all()
        if self.request.GET.get("protection", "") == "false":
            q = q.filter(required_passes=0)
        return get_object_or_404(
            q.select_related(
                "user", "user__spider_info",
            ).prefetch_related("protections"),
            ~models.Q(
                travel_protected__in=travel
            ),
            token=self.kwargs["token"]
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        travel = self.get_travel_for_request().filter(
            login_protection__in=loggedin_active_tprotections
        )
        if self.request.is_owner:
            # request.user is maybe anonymous
            context["content_variants"] = \
                self.usercomponent.user.spider_info.allowed_content.filter(
                    listed_variants_q
                )
            context["machine_variants"] = \
                self.usercomponent.user.spider_info.allowed_content.filter(
                    machine_variants_q
                )
            context["content_variants_used"] = \
                context["content_variants"].filter(
                    ~models.Q(assignedcontent__travel_protected__in=travel),
                    assignedcontent__usercomponent=self.usercomponent
                )
        context["active_features"] = self.usercomponent.features.distinct()
        context["active_listed_features"] = \
            context["active_features"].exclude(
                ctype__contains=VariantType.unlisted.value
            )
        context["is_public_view"] = self.usercomponent.public
        context["has_unlisted"] = self.usercomponent.contents.filter(
            info__contains="\x1eunlisted\x1e"
        ).exclude(travel_protected__in=travel).exists()

        context["remotelink"] = "{}{}?".format(
            context["hostpart"],
            reverse("spider_base:ucontent-list", kwargs={
                "token": self.usercomponent.token
            })
        )
        try:
            # mark_safe required for &
            # especially because of mark_safe DON'T remove int()
            context["remotelink_extra"] = mark_safe("page={}&".format(
                int(self.request.GET.get("page", "1"))
            ))
        except ValueError:
            pass
        return context

    def test_func(self):
        staff_perm = not self.usercomponent.is_index
        if staff_perm:
            staff_perm = "spider_base.view_usercomponent"
        # user token is tested later
        # only test if no token is specified
        if not self.request.GET.get("token") and self.has_special_access(
            user_by_login=True, user_by_token=False,
            staff=staff_perm, superuser=True
        ):
            return True
        # block view on special objects for non user and non superusers
        if self.usercomponent.is_index:
            return False

        # export is only available for user and staff with permission
        minstrength = 0
        if self.scope in ["export"] or self.usercomponent.strength in (4, 9):
            minstrength = 4

        # limit tainted
        ret = self.test_token(minstrength)
        if (
            self.request.auth_token and
            self.request.auth_token.extra.get("taint", False) and
            "referrer" not in self.request.GET
        ):
            return False
        return ret

    def get_paginate_by(self, queryset):
        if self.scope == "export" or "raw" in self.request.GET:
            return None
        return settings.SPIDER_OBJECTS_PER_PAGE

    def render_to_response(self, context):
        if context["scope"] != "export" and "raw" not in self.request.GET:
            return super().render_to_response(context)

        session_dict = {
            "request": self.request,
            "context": context,
            "scope": context["scope"],
            "uc": self.usercomponent,
            "hostpart": context["hostpart"],
            "sourceref": URIRef(context["hostpart"] + self.request.path)
        }
        g = Graph()
        g.namespace_manager.bind("spkc", spkcgraph, replace=True)

        embed = False
        if (
            context["scope"] == "export" or
            self.request.GET.get("raw", "") == "embed"
        ):
            embed = True

        if context["object_list"]:
            p = paginate_stream(
                context["object_list"],
                getattr(
                    settings, "SPIDER_SERIALIZED_PER_PAGE",
                    settings.SPIDER_OBJECTS_PER_PAGE
                ),
                settings.SPIDER_MAX_EMBED_DEPTH
            )
        else:
            # no content, pagination works here only this way
            p = paginate_stream(
                UserComponent.objects.filter(pk=self.usercomponent.pk),
                1,
                1
            )
        page = 1
        try:
            page = int(self.request.GET.get("page", "1"))
        except Exception:
            pass

        if hasattr(self.request, "token_expires"):
            session_dict["expires"] = self.request.token_expires.strftime(
                "%a, %d %b %Y %H:%M:%S %z"
            )

        if page <= 1:
            # "expires" (string) different from "token_expires" (datetime)
            if session_dict.get("expires"):
                add_property(
                    g, "token_expires", ob=session_dict["request"],
                    ref=session_dict["sourceref"]
                )
            for machinec in context.get("machine_variants", []):
                g.add((
                    session_dict["sourceref"],
                    spkcgraph["create:name"],
                    Literal(machinec, datatype=XSD.string)
                ))
            g.add((
                session_dict["sourceref"],
                spkcgraph["scope"],
                Literal(context["scope"], datatype=XSD.string)
            ))
            g.add((
                session_dict["sourceref"],
                spkcgraph["type"],
                Literal("Component", datatype=XSD.string)
            ))
            g.add((
                session_dict["sourceref"],
                spkcgraph["strength"],
                Literal(self.usercomponent.strength, datatype=XSD.integer)
            ))
            if context["referrer"]:
                g.add((
                    session_dict["sourceref"],
                    spkcgraph["referrer"],
                    Literal(context["referrer"], datatype=XSD.anyURI)
                ))
            token = getattr(session_dict["request"], "auth_token", None)
            if token:
                token = token.token
            g.add(
                (
                    session_dict["sourceref"],
                    spkcgraph["action:view"],
                    Literal(
                        merge_get_url(
                            str(session_dict["sourceref"]), token=token
                        ),
                        datatype=XSD.anyURI
                    )
                )
            )
            if context["token_strength"]:
                add_property(
                    g, "token_strength", ref=session_dict["sourceref"],
                    literal=context["token_strength"], datatype=XSD.integer
                )
            add_property(
                g, "intentions", ref=session_dict["sourceref"],
                literal=context["intentions"], datatype=XSD.string,
                iterate=True
            )

        serialize_stream(
            g, p, session_dict,
            page=page,
            embed=embed
        )

        ret = HttpResponse(
            g.serialize(format="turtle"),
            content_type="text/turtle;charset=utf-8"
        )

        if session_dict.get("expires", None):
            ret['X-Token-Expires'] = session_dict["expires"]
        # allow cors requests for raw
        ret["Access-Control-Allow-Origin"] = "*"
        return ret


class ContentAdd(ContentBase, CreateView):
    scope = "add"
    model = ContentVariant
    preserved_GET_parameters = {"token"}

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return self.render_to_response(self.get_context_data())

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        return self.render_to_response(self.get_context_data())

    def form_valid(self, form):
        _ = gettext
        messages.success(
            self.request, _('Content created.')
        )
        return super().form_valid(form)

    def get_queryset(self):
        # use requesting user as base if he can add this type of content
        if self.request.user.is_authenticated:
            return self.request.user.spider_info.allowed_content.filter(
                listed_variants_q |
                models.Q(ctype__contains=VariantType.machine.value)
            )
        else:
            return self.usercomponent.user.spider_info.allowed_content.filter(
                listed_variants_q |
                models.Q(ctype__contains=VariantType.machine.value)
            )

    def test_func(self):
        # test if user and check if user is allowed to create content
        if self.has_special_access(
            user_by_login=True, user_by_token=True, superuser=False
        ):
            if (
                self.request.auth_token and
                self.request.auth_token.extra.get("taint", False)
            ):
                return False
            return True
        return False

    def get_context_data(self, **kwargs):
        kwargs["content_type"] = self.object.installed_class
        kwargs["form"] = self.get_form()
        kwargs["media"] = kwargs["form"].media
        kwargs["active_features"] = self.usercomponent.features.distinct()
        kwargs["active_listed_features"] = \
            kwargs["active_features"].exclude(
                ctype__contains=VariantType.unlisted.value
            )

        kwargs["machine_variants"] = \
            self.usercomponent.user.spider_info.allowed_content.filter(
                machine_variants_q
            )
        return super().get_context_data(**kwargs)

    def get_form(self, allow_data=True):
        assigned = self.object.installed_class.static_create(
            associated_kwargs={
                "usercomponent": self.usercomponent,
                "ctype": self.object
            }
        ).associated
        form_kwargs = {
            "instance": assigned,
            "request": self.request,
            "initial": {
                "usercomponent": self.usercomponent
            }
        }
        if allow_data and self.request.method in ('POST', 'PUT'):
            form_kwargs.update({
                'data': self.request.POST,
                # 'files': self.request.FILES,
            })
        return UserContentForm(**form_kwargs)

    def get_usercomponent(self):
        travel = self.get_travel_for_request().filter(
            login_protection__in=loggedin_active_tprotections
        )
        q = UserComponent.objects.all()
        if self.request.GET.get("protection", "") == "false":
            q = q.filter(required_passes=0)
        return get_object_or_404(
            q.prefetch_related("protections"),
            ~models.Q(travel_protected__in=travel),
            token=self.kwargs["token"],
        )

    def get_object(self, queryset=None):
        if not queryset:
            queryset = self.get_queryset()
        qquery = models.Q(
            name=self.kwargs["type"],
            strength__lte=self.usercomponent.strength
        )
        return get_object_or_404(queryset, qquery)

    def render_to_response(self, context):
        # only true if data
        if context["form"].is_valid():
            ucontent = context["form"].save(commit=False)
        else:
            ucontent = context["form"].instance
        ret = ucontent.content.access(context)

        # return response if content returned response
        if isinstance(ret["content"], HttpResponseBase):
            return ret["content"]
        # redirect if saving worked
        if getattr(ucontent, "id", None):
            assert(ucontent.token)
            assert(ucontent.usercomponent)
            return redirect(
                'spider_base:ucontent-access', permanent=True,
                token=ucontent.token, access="update"
            )
        else:
            assert(not getattr(ucontent.content, "id", None))
        return super().render_to_response(ret)


class ContentAccess(ReferrerMixin, ContentBase, UpdateView):
    scope = "access"
    form_class = UserContentForm
    model = AssignedContent
    rate_limit_group = "spider_static_token_error"
    preserved_GET_parameters = {"token", "search", "id"}

    def dispatch_extra(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.allow_domain_mode = ContentVariant.objects.filter(
            models.Q(feature_for_components=self.usercomponent) |
            models.Q(feature_for_contents=self.object),
            name="DomainMode"
        ).exists()
        # done in get_queryset
        # if getattr(self.request, "auth_token", None):
        #     ids = self.request.auth_token.extra.get("ids", None)
        #     if ids is not None and self.object.id not in ids:
        #         return self.handle_no_permission()
        if "referrer" in self.request.GET:
            self.object_list = self.model.objects.filter(
                pk=self.object.pk
            )
            return self.handle_referrer()

        return None

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        _scope = kwargs["access"]
        # special scopes which should be not available as url parameter
        # raw is also deceptive because view and raw=? = raw scope
        if _scope in _forbidden_scopes:
            raise PermissionDenied("Deceptive scopes")
        self.scope = _scope
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        context = {"form": None}
        if self.scope == "update":
            context["form"] = self.get_form()
        return self.render_to_response(self.get_context_data(**context))

    def post(self, request, *args, **kwargs):
        context = {"form": None}
        # other than update have no form
        if self.scope == "update":
            context["form"] = self.get_form()
            if context["form"].is_valid():
                self.object = context["form"].save(commit=True)
                # use correct usercomponent
                self.usercomponent = context["form"].instance.usercomponent
        return self.render_to_response(self.get_context_data(**context))

    def get_context_data(self, **kwargs):
        kwargs["is_public_view"] = (
            self.usercomponent.public and
            self.scope not in ("add", "update", "raw_update")
        )

        kwargs["machine_variants"] = \
            self.usercomponent.user.spider_info.allowed_content.filter(
                machine_variants_q
            )
        context = super().get_context_data(**kwargs)

        context["remotelink"] = "{}{}?".format(
            context["hostpart"],
            reverse("spider_base:ucontent-access", kwargs={
                "token": self.object.token,
                "access": "view"
            })
        )
        if self.scope == "update":
            context["active_features"] = self.usercomponent.features.distinct()
        else:
            context["active_features"] = ContentVariant.objects.filter(
                models.Q(feature_for_contents=self.object) |
                models.Q(feature_for_components=self.usercomponent)
            ).distinct()
        context["active_listed_features"] = \
            context["active_features"].exclude(
                ctype__contains=VariantType.unlisted.value
            )
        return context

    def get_form_success_kwargs(self):
        """Return the keyword arguments for instantiating the form."""
        return {
            'initial': self.get_initial(),
            'instance': self.object,
            'request': self.request,
            'prefix': self.get_prefix()
        }

    def get_form_kwargs(self):
        """Return the keyword arguments for instantiating the form."""
        ret = super().get_form_kwargs()
        ret["request"] = self.request
        return ret

    def test_func(self):
        # give user and staff the ability to update Content
        # except it is index, in this case only the user can update
        # reason: admins could be tricked into malicious updates
        # for index the same reason as for add
        is_index = self.usercomponent.is_index
        staff_perm = not is_index
        if staff_perm:
            staff_perm = "spider_base.view_assignedcontent"
            if self.scope in {"update", "raw_update"}:
                staff_perm = "spider_base.update_assignedcontent"
        # user token is tested later
        # only test if no token is specified
        if not self.request.GET.get("token") and self.has_special_access(
            staff=staff_perm, superuser=(not is_index),
            user_by_token=False, user_by_login=True
        ):
            return True
        if is_index:
            return False
        minstrength = 0
        if (
            self.scope in {"update", "raw_update", "export"} or
            self.usercomponent.strength in (4, 9)
        ):
            minstrength = 4
        ret = self.test_token(minstrength)
        if (
            self.request.auth_token and
            self.request.auth_token.extra.get("taint", False) and
            "referrer" not in self.request.GET
        ):
            return False
        return ret

    def get_usercomponent(self):
        travel = self.get_travel_for_request().filter(
            login_protection__in=loggedin_active_tprotections
        )
        q = UserComponent.objects.all()
        if self.request.GET.get("protection", "") == "false":
            q = q.filter(required_passes=0)
        return get_object_or_404(
            q.prefetch_related("protections"),
            ~models.Q(
                travel_protected__in=travel
            ),
            contents__token=self.kwargs["token"]
        )

    def get_object(self, queryset=None):
        # can bypass idlist and searchlist with own queryset arg
        if not queryset:
            queryset = self.get_queryset()
        # doesn't matter if it is same user, lazy
        travel = self.get_travel_for_request().filter(
            login_protection__in=loggedin_active_tprotections
        )

        # required for next/previous token
        queryset = queryset.select_related(
            "usercomponent", "usercomponent__user",
            "usercomponent__user__spider_info"
        ).filter(usercomponent=self.usercomponent).order_by("priority", "id")
        ob = get_object_or_404(
            queryset,
            (
                ~models.Q(travel_protected__in=travel) |
                models.Q(
                    ctype__name__in={"SelfProtection", "TravelProtection"}
                )
            ),
            token=self.kwargs["token"]
        )
        queryset2 = queryset.exclude(
            ~models.Q(id=ob.id),
            info__contains="\x1eunlisted\x1e",
        )
        ob.previous_object = prev_in_order(ob, queryset2)
        ob.next_object = next_in_order(ob, queryset2)
        # for receiving updates without refresh_from_db
        ob.content.associated_obj = ob
        return ob

    def render_to_response(self, context):
        # context is updated and used outside!!
        ret = self.object.content.access(context)

        # allow contents to redirect from update
        #   (e.g. if user should not know static token)
        if isinstance(ret["content"], HttpResponseBase):
            return ret["content"]
        if self.scope == "update":
            # token changed => path has changed
            if self.object.token != self.kwargs["token"]:
                return redirect(
                    'spider_base:ucontent-access', permanent=True,
                    token=self.object.token, access="update"
                )

            if ret["form"].is_valid():
                ret["form"] = self.get_form_class()(
                    **self.get_form_success_kwargs()
                )

        return super().render_to_response(ret)


class ContentDelete(EntityDeletionMixin, DeleteView):
    model = AssignedContent
    usercomponent = None
    preserved_GET_parameters = {"token", "search", "id"}

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.usercomponent = self.get_usercomponent()
        try:
            return super().dispatch(request, *args, **kwargs)
        except PermissionDenied as exc:
            if request.is_staff:
                raise exc
            # elsewise disguise
            raise Http404()

    def delete(self, request, *args, **kwargs):
        _ = gettext
        try:
            return super().delete(request, *args, **kwargs)
        except ProtectedError as exc:
            # hide if travelprotected causes ProtectedError
            travel = self.get_travel_for_request()
            travel = travel.filter(
                models.Q(
                    protect_contents__id__in=exc.protected_objects.values_list(
                        "associated_rel__id", flat=True
                    )
                ) |
                models.Q(
                    protect_components__id__in=exc.protected_objects.values_list(  # noqa: E501
                        "associated_rel__usercomponent__id", flat=True
                    )
                ),
                login_protection__in=loggedin_active_tprotections
            )
            obj = travel.order_by("-start").filter(stop__isnull=True).first()
            if not obj:
                try:
                    obj = travel.latest("stop")
                except travel.model.DoesNotExist:
                    pass
            if obj:
                obj.protect_contents.add(self.object)
                return HttpResponseRedirect(self.get_success_url())
            else:
                messages.error(
                    self.request,
                    mark_safe(
                        _(
                            'Could not delete content dependencies:<br/>{}.'
                        ).format(
                            ",<br/>".join(map(
                                lambda x: escape(repr(x)),
                                exc.protected_objects
                            ))
                        )
                    )
                )
                return self.get(request, *args, **kwargs)

    def form_valid(self, form):
        _ = gettext
        messages.success(
            self.request, _('Content deleted.')
        )
        return super().form_valid(form)

    def get_context_data(self, **kwargs):
        kwargs["uc"] = self.usercomponent
        return super().get_context_data(**kwargs)

    def get_success_url(self):
        return reverse(
            "spider_base:ucontent-list", kwargs={
                "token": self.usercomponent.token,
            }
        )

    def test_func(self):
        # can do it over admin panel but for convenience and \
        # allowing users to cancel (in case of timeouts)
        staff_perm = not self.usercomponent.is_index
        if staff_perm:
            staff_perm = "spider_base.delete_assignedcontent"
        ret = self.has_special_access(
            user_by_login=True, user_by_token=True,
            staff=staff_perm, superuser=True
        )
        if (
            self.request.auth_token and
            self.request.auth_token.extra.get("taint", False) and
            "referrer" not in self.request.GET
        ):
            return False
        return ret

    def get_required_timedelta(self):
        _time = self.object.content.deletion_period
        if _time:
            _time = timedelta(seconds=_time)
        else:
            _time = timedelta(seconds=0)
        return _time

    def get_usercomponent(self):
        return self.object.usercomponent

    def get_object(self, queryset=None):
        if not queryset:
            queryset = self.get_queryset()
        travel = self.get_travel_for_request().filter(
            login_protection__in=loggedin_active_tprotections
        )
        return get_object_or_404(
            queryset,
            ~(
                models.Q(travel_protected__in=travel) |
                models.Q(usercomponent__travel_protected__in=travel)
            ),
            token=self.kwargs["token"]
        )


class TravelProtectionManagement(UserTestMixin, UpdateView):
    model = TravelProtection
    template_name = "spider_base/travelprotection.html"
    form_class = TravelProtectionManagementForm

    def dispatch(self, request, *args, **kwargs):
        _ = gettext
        try:
            self.object = self.get_object()
            if not self.object.active:
                messages.success(
                    self.request, _('Protection not active')
                )
                return redirect("home")
            return super().dispatch(request, *args, **kwargs)
        except Http404:
            return get_settings_func(
                "SPIDER_RATELIMIT_FUNC",
                "spkcspider.apps.spider.functions.rate_limit_default"
            )(self, request)

    def get_object(self, queryset=None):
        if queryset is None:
            queryset = self.get_queryset()

        url = self.request.GET.get("url")
        if url:
            url = static_token_matcher.match(url)
        if not url:
            raise Http404()
        url = url.groupdict()
        return get_object_or_404(
            queryset,
            associated_rel__token=url["static_token"]
        )

    def get(self, request, *args, **kwargs):
        return self.render_to_response(self.get_context_data())

    def post(self, request, *args, **kwargs):
        form = self.get_form()
        if form.is_valid():
            return self.form_valid(form)
        else:
            return self.form_invalid(form)

    def render_to_response(self, context):
        if not (
            self.object.associated.getflag("anonymous_deactivation") or
            #  not implemented yet
            self.object.associated.getflag("anonymous_trigger")
        ):
            return HttpResponseRedirect(
                redirect_to=self.object.associated.get_absolute_url("update")
            )
        return super().render_to_response(context)

    def form_valid(self, form):
        _ = gettext
        self.object.active = False
        self.object._anonymous_deactivation = \
            self.object.associated.getflag("anonymous_deactivation")
        self.object._encoded_pwhashes = "".join(
            map(
                lambda x: "pwhash={}\x1e".format(x),
                self.object.associated.getlist("pwhash", 20)
            )
        )
        self.object.clean()
        self.object.save(update_fields=["active"])
        messages.success(
            self.request, _('{} disabled').format(self.object.associated.ctype)
        )
        return redirect("home")

    def options(self, request, *args, **kwargs):
        ret = super().options()
        ret["Access-Control-Allow-Origin"] = "*"
        ret["Access-Control-Allow-Methods"] = "POST, GET, OPTIONS"
        return ret

    def test_func(self):
        return True



__all__ = ["SpiderWebCfgConfig"]

from django.apps import AppConfig


class SpiderWebCfgConfig(AppConfig):
    name = 'spkcspider.apps.spider_webcfg'
    label = 'spider_webcfg'
    spider_url_path = 'webcfg/'
    verbose_name = 'spkcspider WebConfig'



__all__ = ["ReferrerMixin"]


import hashlib
import logging
from urllib.parse import quote_plus

from django.http import (
    HttpResponseRedirect, HttpResponseServerError, HttpResponse
)
from django.db import transaction
from django.db.models import Q
from django.conf import settings
from django.utils.translation import gettext
from django.test import Client
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.forms.widgets import Media

import ratelimit

import requests

from ..helpers import (
    merge_get_url, get_requests_params, get_settings_func, get_hashob
)
from ..constants import TokenCreationError
from ..conf import VALID_INTENTIONS, VALID_SUB_INTENTIONS
from ..models import AuthToken, ReferrerObject

logger = logging.getLogger(__name__)
_extra = '' if settings.DEBUG else '.min'


class ReferrerMixin(object):
    allow_domain_mode = False

    def get_context_data(self, **kwargs):
        kwargs["token_strength"] = None
        # will be overwritten in referring path so there is no interference
        kwargs["referrer"] = None
        kwargs["intentions"] = set()
        if self.request.auth_token:
            kwargs["referrer"] = self.request.auth_token.referrer
            kwargs["token_strength"] = self.request.auth_token.extra.get(
                "strength", None
            )
            kwargs["intentions"].update(self.request.auth_token.extra.get(
                "intentions", []
            ))
        return super().get_context_data(**kwargs)

    def test_token(self, minstrength, force_token=False, taint=False):
        if "intention" in self.request.GET or "referrer" in self.request.GET:
            # validate early, before auth
            intentions = set(self.request.GET.getlist("intention"))
            if not VALID_INTENTIONS.issuperset(
                intentions
            ):
                return HttpResponse(
                    "invalid intentions", status=400
                )
            if "domain" in intentions:

                if not self.clean_domain_upgrade(
                    {"intentions": intentions},
                    False
                ):
                    return HttpResponse(
                        "invalid domain upgrade", status=400
                    )
                # requires token
                force_token = True
            else:
                # can be either domain or auth to not have taint flag
                if "auth" not in intentions:
                    taint = True
                # maximal one main intention
                if len(intentions.difference(VALID_SUB_INTENTIONS)) > 1:
                    return HttpResponse(
                        "invalid intentions", status=400
                    )
                minstrength = 4
        return super().test_token(minstrength, force_token, taint)

    def refer_with_post(self, context, token):
        # application/x-www-form-urlencoded is best here,
        # for beeing compatible to most webservers
        # client side rdf is no problem
        # NOTE: csrf must be disabled or use csrf token from GET,
        #       here is no way to know the token value

        h = hashlib.sha256(context["referrer"].encode("utf8")).hexdigest()

        def h_fun(*a):
            return h
        # rate limit on errors
        if ratelimit.get_ratelimit(
            request=self.request,
            group="refer_with_post.refer_with_post",
            key=h_fun,
            rate=settings.SPIDER_DOMAIN_ERROR_RATE,
            inc=False
        )["request_limit"] > 0:
            return HttpResponseRedirect(
                redirect_to=merge_get_url(
                    context["referrer"],
                    status="post_failed",
                    error="error_rate_limit"
                )
            )
        d = {
            "token": token.token,
            "hash_algorithm": settings.SPIDER_HASH_ALGORITHM.name,
            "action": context["action"],
        }
        if context["payload"] is not None:
            d["payload"] = context["payload"]
        params, can_inline = get_requests_params(context["referrer"])
        if can_inline:
            response = Client().post(
                context["referrer"],
                data=d,
                Connection="close",
                Referer=merge_get_url(
                    "%s%s" % (
                        context["hostpart"],
                        self.request.path
                    )
                    # sending full url not required anymore, payload
                )
            )
            if response.status_code != 200:
                return HttpResponseRedirect(
                    redirect_to=merge_get_url(
                        context["referrer"],
                        status="post_failed",
                        error="other"
                    )
                )
        else:
            try:
                with requests.post(
                    context["referrer"],
                    data=d,
                    headers={
                        "Referer": merge_get_url(
                            "%s%s" % (
                                context["hostpart"],
                                self.request.path
                            )
                            # sending full url not required anymore, payload
                        ),
                        "Connection": "close"
                    },
                    **params
                ) as resp:
                    resp.raise_for_status()
            except requests.exceptions.SSLError as exc:
                logger.info(
                    "referrer: \"%s\" has a broken ssl configuration",
                    context["referrer"], exc_info=exc
                )
                return HttpResponseRedirect(
                    redirect_to=merge_get_url(
                        context["referrer"],
                        status="post_failed",
                        error="ssl"
                    )
                )
            except Exception as exc:
                apply_error_limit = False
                if isinstance(
                    exc, (
                        requests.exceptions.ConnectionError,
                        requests.exceptions.Timeout
                    )
                ):
                    apply_error_limit = True
                elif (
                    isinstance(exc, requests.exceptions.HTTPError) and
                    exc.response.status_code >= 500
                ):
                    apply_error_limit = True
                if apply_error_limit:
                    ratelimit.get_ratelimit(
                        request=self.request,
                        group="refer_with_post",
                        key=h_fun,
                        rate=settings.SPIDER_DOMAIN_ERROR_RATE,
                        inc=True
                    )
                logger.info(
                    "post failed: \"%s\" failed",
                    context["referrer"], exc_info=exc
                )
                return HttpResponseRedirect(
                    redirect_to=merge_get_url(
                        context["referrer"],
                        status="post_failed",
                        error="other"
                    )
                )
        context["post_success"] = True
        h = get_hashob()
        h.update(token.token.encode("utf-8", "ignore"))
        return HttpResponseRedirect(
            redirect_to=merge_get_url(
                context["referrer"],
                status="success",
                hash=h.finalize().hex()
            )
        )

    def refer_with_get(self, context, token):
        return HttpResponseRedirect(
            redirect_to=merge_get_url(
                context["referrer"],
                token=token.token,
                payload=context["payload"]
            )
        )

    def clean_domain_upgrade(self, context, token):
        if "referrer" not in self.request.GET:
            return False
        # domain mode must be used alone
        if len(context["intentions"]) > 1:
            return False
        if not context["intentions"].issubset(VALID_INTENTIONS):
            return False

        if not getattr(self.request, "_clean_domain_upgrade_checked", False):
            if ratelimit.get_ratelimit(
                request=self.request,
                group="clean_domain_upgrade",
                key=("get", {"IP": True, "USER": True}),
                rate=settings.SPIDER_DOMAIN_UPDATE_RATE,
                inc=True
            )["request_limit"] > 0:
                return False

            setattr(self.request, "_clean_domain_upgrade_checked", True)

        # False for really no token
        if token is False:
            return True
        if not token or token.extra.get("strength", 0) >= 10:
            return False
        return True

    def clean_refer_intentions(self, context, token=None):
        # Only owner can use other intentions than domain
        if not self.request.is_owner:
            return False
        # Second error: invalid intentions
        #  this is the second time the validation will be executed
        #    in case test_token path is used
        #  this is the first time the validation will be executed
        #    in case has_special_access path is used
        if not context["intentions"].issubset(VALID_INTENTIONS):
            return False

        # auth is only for self.requesting component auth
        if "auth" in context["intentions"]:
            return False
        # maximal one main intention
        if len(context["intentions"].difference(VALID_SUB_INTENTIONS)) > 1:
            return False
        # "persist" or default can be serverless other intentions not
        #  this way rogue client based attacks are prevented
        if "persist" in context["intentions"]:
            if not self.usercomponent.features.filter(
                name="Persistence"
            ).exists():
                return False
        else:
            if context["is_serverless"] and len(context["intentions"]) != 1:
                return False

        if not token:
            return True

        ####### with token ########  # noqa: 266E
        # cannot add sl intention to existing intentions
        if "sl" in context["intentions"].difference(
            token.extra.get("intentions", ["sl"])
        ):
            return False
        if "persist" in context["intentions"]:
            # set persist = true, (false=-1)
            token.persist = 0
            # if possible, pin to anchor
            if self.usercomponent.primary_anchor:
                token.persist = self.usercomponent.primary_anchor.id
        else:
            token.persist = -1

        if not token.extra.get("initial_referrer_url", None):
            token.extra["initial_referrer_url"] = "{}://{}{}".format(
                self.request.scheme,
                self.request.get_host(),
                self.request.path
            )
        return True

    def handle_domain_auth(self, context, token):
        assert token or self.request.is_special_user, \
            "special user and no token"

        context.setdefault("payload", None)
        context.setdefault("action", "create")
        if not self.allow_domain_mode:
            return HttpResponse(
                status=400,
                content='domain mode disallowed'
            )
        if not self.clean_domain_upgrade(context, token):
            return HttpResponse(
                status=400,
                content='Invalid token'
            )
        token.initialize_token()
        token.referrer = ReferrerObject.objects.get_or_create(
            url=context["referrer"]
        )[0]
        token.extra["intentions"] = list(context["intentions"])
        try:
            token.save()
        except TokenCreationError:
            logger.exception("Token creation failed")
            return HttpResponseServerError(
                "Token creation failed, try again"
            )
        context["post_success"] = False
        ret = self.refer_with_post(context, token)
        if not context["post_success"]:
            token.delete()
        return ret

    def handle_referrer_request(
        self, context, token, keep=False, dontact=False, no_oldtoken=False
    ):
        """
            no_oldtoken: don't use old token for calculating:
                           old_ids, old_search
                         (performance and manual old_* possible)
            dontact: dont act if probing results fails
        """
        _ = gettext
        context.setdefault("action", "create")
        context.setdefault("model", self.model)
        context.setdefault("payload", None)
        context.setdefault("ids", set())
        context.setdefault("filter", set())
        context.setdefault("old_ids", set())
        context.setdefault("old_search", set())
        context["is_serverless"] = "sl" in context["intentions"]
        if keep is True:
            context["ids"].update(token.extra.get("ids", []))
            context["search"].update(token.extra.get("filter", []))

        action = self.request.POST.get("action", None)
        if context["action"].endswith("invalid"):
            action = context["action"]
        if action == "confirm":
            newtoken = None
            # if persist try to find old token
            if "persist" in context["intentions"]:
                oldtoken = AuthToken.objects.filter(
                    Q(persist__gte=0, usercomponent=token.usercomponent),
                    referrer__url=context["referrer"]
                ).first()
                if oldtoken:
                    newtoken = token
                    token = oldtoken
            # either reuse persistent token with auth token tokenstring
            # or just reuse auth token
            if newtoken:
                # steal token value
                token.token = newtoken.token

            # set to zero as prot_strength can elevate perms
            token.extra["taint"] = False
            token.extra["prot_strength"] = 0
            token.extra["intentions"] = list(context["intentions"])
            token.extra.pop("request_referrer", None)
            token.extra.pop("request_intentions", None)
            token.extra.pop("request_search", None)

            if not self.clean_refer_intentions(context, token):
                return HttpResponseRedirect(
                    redirect_to=merge_get_url(
                        context["referrer"],
                        error="intentions_incorrect"
                    )
                )

            token.extra["search"] = list(context["search"])
            if "live" in context["intentions"]:
                token.extra.pop("ids", None)
            else:
                token.extra["ids"] = list(context["ids"])
            token.referrer = ReferrerObject.objects.get_or_create(
                url=context["referrer"]
            )[0]
            # after cleanup, save
            try:
                with transaction.atomic():
                    # must be done here, elsewise other token can (unlikely)
                    # take token as it is free for a short time, better be safe
                    if newtoken:
                        newtoken.delete()
                    token.save()
            except TokenCreationError:
                logger.exception("Token creation failed")
                return HttpResponseServerError(
                    _("Token creation failed, try again")
                )

            if context["is_serverless"]:
                context["post_success"] = True
                ret = self.refer_with_get(context, token)
            else:
                context["post_success"] = False
                ret = self.refer_with_post(context, token)
            if dontact:
                return ret
            if not context["post_success"]:
                if newtoken:
                    logger.warning(
                        "Updating persisting token failed"
                    )
                else:
                    token.delete()
            return ret

        elif action == "cancel":
            if not dontact:
                token.delete()
            return HttpResponseRedirect(
                redirect_to=merge_get_url(
                    context["referrer"],
                    status="canceled",
                    payload=context["payload"]
                )
            )
        else:
            if (
                no_oldtoken and
                "persist" in context["intentions"]
            ):
                oldtoken = AuthToken.objects.filter(
                    Q(persist__gte=0, usercomponent=token.usercomponent),
                    referrer__url=context["referrer"],
                ).first()

                if oldtoken:
                    context["old_ids"].update(oldtoken.extra.get("ids", []))
                    context["old_search"].update(oldtoken.extra.get(
                        "filter", []
                    ))

            if not self.clean_refer_intentions(context, token):
                return HttpResponse(
                    status=400,
                    content=_('Error: intentions incorrect')
                )
            context["object_list"] = context["model"].objects.filter(
                id__in=context["ids"]
            )
            # remove other media
            context["media"] = Media(
                css={
                    "all": [
                        'node_modules/selectize/dist/css/selectize.default.css'
                    ]
                },
                js=[
                    'node_modules/jquery/dist/jquery%s.js' % _extra,
                    'node_modules/selectize/dist/js/standalone/selectize%s.js' % _extra  # noqa: E501
                ]
            )
            return self.response_class(
                request=self.request,
                template=self.get_referrer_template_names(),
                context=context,
                using=self.template_engine,
                content_type=self.content_type
            )

    def handle_referrer(self):
        _ = gettext
        if (
            self.request.user != self.usercomponent.user and
            not self.request.auth_token
        ):
            if self.request.user.is_authenticated:
                return self.handle_no_permission()
            return HttpResponseRedirect(
                redirect_to="{}?{}={}".format(
                    self.get_login_url(),
                    REDIRECT_FIELD_NAME,
                    quote_plus(
                        merge_get_url(
                            self.request.build_absolute_uri(),
                            token=None
                        )
                    )
                )
            )

        context = self.get_context_data()
        context["action"] = "create"
        context["intentions"] = set(self.request.GET.getlist("intention"))
        if "referrer" in self.request.POST:
            context["referrer"] = merge_get_url(self.request.POST["referrer"])
            if not get_settings_func(
                "SPIDER_URL_VALIDATOR",
                "spkcspider.apps.spider.functions.validate_url_default"
            )(context["referrer"], self):
                context["action"] = "referrer_invalid"
        else:
            context["referrer"] = merge_get_url(self.request.GET["referrer"])
            if not get_settings_func(
                "SPIDER_URL_VALIDATOR",
                "spkcspider.apps.spider.functions.validate_url_default"
            )(context["referrer"], self):
                return HttpResponse(
                    status=400,
                    content=_('Insecure url: %(url)s') % {
                        "url": context["referrer"]
                    }
                )
        context["payload"] = self.request.GET.get("payload", None)
        token = self.request.auth_token
        if not token:
            token = AuthToken(
                usercomponent=self.usercomponent,
                extra={
                    "strength": 10,
                    "taint": False
                }
            )
        if "domain" in context["intentions"]:
            return self.handle_domain_auth(context, token)
        else:
            context["ids"] = set(self.object_list.values_list("id", flat=True))
            context["search"] = set(self.request.POST.getlist("search"))
            return self.handle_referrer_request(
                context, token
            )

    def get_referrer_template_names(self):
        return "spider_base/protections/referring.html"



from ._core import *  # noqa: F403, F401
from ._components import *  # noqa: F403, F401
from ._contents import *  # noqa: F403, F401
from ._referrer import *  # noqa: F403, F401
from ._tokens import *  # noqa: F403, F401



__all__ = ["WebConfigForm"]

from django import forms
from .models import WebConfig


class WebConfigForm(forms.ModelForm):
    creation_url = forms.URLField(disabled=True, required=False)

    class Meta:
        model = WebConfig
        fields = ['config']
        widgets = {
            "config": forms.Textarea()
        }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.initial["creation_url"] = \
            self.instance.associated.attached_to_token.referrer.url



__all__ = [
    "SelectizeWidget", "OpenChoiceWidget", "StateButtonWidget",
    "TrumbowygWidget", "HTMLWidget", "SubSectionStartWidget",
    "SubSectionStopWidget", "ListWidget", "RomeDatetimePickerWidget"
]


from django.forms import widgets
from django.conf import settings
from django.utils.translation import get_language, gettext_lazy as _


_extra = '' if settings.DEBUG else '.min'


class StateButtonWidget(widgets.CheckboxInput):
    template_name = 'spider_base/forms/widgets/statebutton.html'
    # ugly as hell

    class Media:
        css = {
            'all': [
                'spider_base/statebutton.css'
            ]
        }


class ListWidget(widgets.SelectMultiple):
    template_name = 'spider_base/forms/widgets/wrapped_select.html'
    allow_multiple_selected = True

    class Media:
        js = [
            'node_modules/@json-editor/json-editor/dist/jsoneditor%s.js' % _extra,  # noqa:E501
            'spider_base/ListEditorWidget.js'
        ]

    def __init__(
        self, *, attrs=None, wrapper_attrs=None,
        format_type="text", item_label=_("Item"), **kwargs
    ):
        if not attrs:
            attrs = {"class": ""}
        if not wrapper_attrs:
            wrapper_attrs = {}
        attrs.setdefault("class", "")
        attrs["class"] += " SpiderListTarget"
        attrs["data-format_type"] = format_type
        # don't access them as they are lazy evaluated
        attrs["data-item_label"] = item_label
        self.wrapper_attrs = wrapper_attrs.copy()
        super().__init__(attrs=attrs, **kwargs)

    def __deepcopy__(self, memo):
        obj = super().__deepcopy__(memo)
        obj.wrapper_attrs = self.wrapper_attrs.copy()
        return obj

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        context['widget']['wrapper_attrs'] = self.wrapper_attrs
        context['widget']['wrapper_attrs']["id"] = "{}_inner_wrapper".format(
            context['widget']['attrs']["id"]
        )
        return context

    def optgroups(self, name, value, attrs=None):
        """Return a list of optgroups for this widget."""
        groups = []

        for index, option_value in enumerate(value or []):
            if option_value is None:
                option_value = ''
            # selected = True
            groups.append((
                None,
                [self.create_option(
                    name, option_value, option_value, True, index,
                    subindex=None, attrs=attrs,
                )],
                index
            ))
        return groups


class HTMLWidget(widgets.Input):
    template_name = 'spider_base/forms/widgets/info.html'
    input_type = "html_widget"

    def __init__(self, *, template_name=None, **kwargs):
        if template_name:
            self.template_name = template_name
        super().__init__(**kwargs)

    def use_required_attribute(self, initial):
        return False

    def format_value(self, value):
        return None

    def value_omitted_from_data(self, data, files, name):
        return None


class RomeDatetimePickerWidget(widgets.DateTimeInput):
    anchor_class = "RomeDatetimeTarget"

    def __init__(self, *, attrs=None, **kwargs):
        if not attrs:
            attrs = {"class": ""}
        attrs.setdefault("class", "")
        attrs["class"] += " %s" % self.anchor_class
        super().__init__(attrs=attrs, **kwargs)

    class Media:
        css = {
            "all": [
                "node_modules/@bevacqua/rome/dist/rome%s.css" % _extra
            ]
        }
        js = [
            'node_modules/@bevacqua/rome/dist/rome%s.js' % _extra,  # noqa:E501
            'spider_base/RomePickerWidget.js'
        ]


class SubSectionStartWidget(HTMLWidget):
    input_type = "start_pseudo"
    template_name = 'spider_base/forms/widgets/subsectionstart.html'
    label = None

    def __init__(self, *, label=None, **kwargs):
        self.label = label
        super().__init__(**kwargs)

    def get_context(self, name, value, attrs):
        ret = super().get_context(name, value, attrs)
        ret["label"] = self.label
        return ret


class SubSectionStopWidget(HTMLWidget):
    input_type = "stop_pseudo"
    template_name = 'spider_base/forms/widgets/subsectionstop.html'


class SelectizeWidget(widgets.Select):
    anchor_class = "SelectizeWidgetTarget"

    class Media:
        css = {
            'all': [
                'node_modules/selectize/dist/css/selectize.default.css'
            ]
        }
        js = [
            'node_modules/jquery/dist/jquery%s.js' % _extra,
            'node_modules/selectize/dist/js/standalone/selectize%s.js' % _extra,  # noqa: E501
            'spider_base/SelectizeWidget.js'
        ]

    def __init__(self, allow_multiple_selected, *, attrs=None, **kwargs):
        self.allow_multiple_selected = allow_multiple_selected
        if not attrs:
            attrs = {"class": ""}
        attrs.setdefault("style", "width:100%")
        attrs.setdefault("class", "")
        attrs["class"] += " %s" % self.anchor_class
        super().__init__(attrs=attrs, **kwargs)


class OpenChoiceWidget(widgets.Select):
    anchor_class = "OpenChoiceTarget"

    class Media:
        css = {
            'all': [
                'node_modules/selectize/dist/css/selectize.default.css'
            ]
        }
        js = [
            'node_modules/jquery/dist/jquery%s.js' % _extra,
            'node_modules/selectize/dist/js/standalone/selectize%s.js' % _extra,  # noqa: E501
            'spider_base/OpenChoiceWidget.js'
        ]

    def __init__(self, allow_multiple_selected, *, attrs=None, **kwargs):
        self.allow_multiple_selected = allow_multiple_selected
        if not attrs:
            attrs = {"class": ""}
        attrs.setdefault("style", "width:100%")
        attrs.setdefault("class", "")
        attrs["class"] += " %s" % self.anchor_class
        super().__init__(attrs=attrs, **kwargs)

    def optgroups(self, name, value, attrs=None):
        """Return a list of optgroups for this widget."""
        groups = []
        groups2 = []
        has_selected = False

        choices = list(self.choices)
        value_set = set(value)
        # optimization
        if isinstance(value, set):
            value = value_set

        lastindex = 0
        for index, (option_value, option_label) in enumerate(choices):
            if option_value is None:
                option_value = ''

            subgroup = []
            if isinstance(option_label, (list, tuple)):
                group_name = option_value
                subindex = 0
                choices = option_label
            else:
                group_name = None
                subindex = None
                choices = [(option_value, option_label)]
            groups.append((group_name, subgroup, index))

            for subvalue, sublabel in choices:
                selected = False
                if str(subvalue) in value_set:
                    value_set.remove(str(subvalue))
                    selected = True
                selected = (
                    selected and
                    (not has_selected or self.allow_multiple_selected)
                )
                has_selected |= selected
                subgroup.append(self.create_option(
                    name, subvalue, sublabel, selected, index,
                    subindex=subindex, attrs=attrs,
                ))
                if subindex is not None:
                    subindex += 1
            lastindex = index

        for index, option_value in enumerate(value, lastindex):
            if option_value not in value_set:
                continue
            if option_value is None:
                option_value = ''

            subgroup = [self.create_option(
                name, option_value, option_value,
                (not has_selected or self.allow_multiple_selected), index,
                subindex=None, attrs=attrs,
            )]
            groups2.append((None, subgroup, index))

        return groups2 + groups


class TrumbowygWidget(widgets.Textarea):
    template_name = 'spider_base/forms/widgets/wrapped_textarea.html'

    _media = widgets.Media(
        css={
            'all': [
                'node_modules/trumbowyg/dist/ui/trumbowyg.min.css',
                'node_modules/trumbowyg/dist/plugins/emoji/ui/trumbowyg.emoji.css',  # noqa: E501
                'node_modules/trumbowyg/dist/plugins/colors/ui/trumbowyg.colors.css',  # noqa: E501
                # 'node_modules/trumbowyg/dist/plugins/history/ui/trumbowyg.history.css'  # noqa: E501
                # 'node_modules/jquery-resizable/resizable.css',
            ]
        },
        js=[
            'node_modules/jquery/dist/jquery%s.js' % _extra,
            # 'node_modules/jquery-resizable/resizable.js',
            'node_modules/trumbowyg/dist/trumbowyg%s.js' % _extra,
            'node_modules/trumbowyg/dist/plugins/emoji/trumbowyg.emoji.min.js',
            'node_modules/trumbowyg/dist/plugins/preformatted/trumbowyg.preformatted.min.js',  # noqa: E501
            'node_modules/trumbowyg/dist/plugins/insertaudio/trumbowyg.insertaudio.min.js',  # noqa: E501
            'node_modules/trumbowyg/dist/plugins/pasteimage/trumbowyg.pasteimage.min.js',  # noqa: E501
            'node_modules/trumbowyg/dist/plugins/lineheight/trumbowyg.lineheight.min.js',  # noqa: E501
            'node_modules/trumbowyg/dist/plugins/base64/trumbowyg.base64.min.js',  # noqa: E501
            'node_modules/trumbowyg/dist/plugins/history/trumbowyg.history.min.js',  # noqa: E501
            'node_modules/trumbowyg/dist/plugins/colors/trumbowyg.colors.min.js',  # noqa: E501
            # 'node_modules/trumbowyg/dist/plugins/resizimg/trumbowyg.resizimg.js',  # noqa: E501
            'node_modules/trumbowyg/dist/plugins/table/trumbowyg.table.min.js',
            'spider_base/trumbowygWidget.js',
        ]
    )
    @property
    def media(self):
        if not self.is_localized:
            return self._media
        lang = get_language()
        if not lang or lang.startswith("en"):
            return self._media
        lang = lang.replace("-", "_")
        return widgets.Media(
            css={
                'all': [
                    'node_modules/trumbowyg/dist/ui/trumbowyg.min.css',
                    'node_modules/trumbowyg/dist/plugins/emoji/ui/trumbowyg.emoji.css',  # noqa: E501
                    'node_modules/trumbowyg/dist/plugins/colors/ui/trumbowyg.colors.css',  # noqa: E501
                    # 'node_modules/trumbowyg/dist/plugins/history/ui/trumbowyg.history.css'  # noqa: E501
                    # 'node_modules/jquery-resizable/resizable.css',
                ]
            },
            js=[
                'node_modules/jquery/dist/jquery%s.js' % _extra,
                # 'node_modules/jquery-resizable/resizable.js',
                'node_modules/trumbowyg/dist/trumbowyg%s.js' % _extra,
                'node_modules/trumbowyg/dist/plugins/emoji/trumbowyg.emoji.min.js',  # noqa: E501
                'node_modules/trumbowyg/dist/plugins/preformatted/trumbowyg.preformatted.min.js',  # noqa: E501
                'node_modules/trumbowyg/dist/plugins/insertaudio/trumbowyg.insertaudio.min.js',  # noqa: E501
                'node_modules/trumbowyg/dist/plugins/pasteimage/trumbowyg.pasteimage.min.js',  # noqa: E501
                'node_modules/trumbowyg/dist/plugins/lineheight/trumbowyg.lineheight.min.js',  # noqa: E501
                'node_modules/trumbowyg/dist/plugins/base64/trumbowyg.base64.min.js',  # noqa: E501
                'node_modules/trumbowyg/dist/plugins/history/trumbowyg.history.min.js',  # noqa: E501
                'node_modules/trumbowyg/dist/plugins/colors/trumbowyg.colors.min.js',  # noqa: E501
                # 'node_modules/trumbowyg/dist/plugins/resizimg/trumbowyg.resizimg.js',  # noqa: E501
                'node_modules/trumbowyg/dist/plugins/table/trumbowyg.table.min.js',  # noqa: E501
                'spider_base/trumbowygWidget.js',
                'node_modules/trumbowyg/dist/langs/{}.min.js'.format(
                    lang
                )
            ]
        )

    def __init__(self, *, attrs=None, wrapper_attrs=None, **kwargs):
        if not attrs:
            attrs = {"class": ""}
        if not wrapper_attrs:
            wrapper_attrs = {"class": ""}
        attrs.setdefault("class", "")
        wrapper_attrs.setdefault("class", "")
        attrs["class"] += " TrumbowygTarget"
        wrapper_attrs["class"] += " TrumbowygTargetWrapper"
        self.wrapper_attrs = wrapper_attrs.copy()
        super().__init__(attrs=attrs, **kwargs)

    def __deepcopy__(self, memo):
        obj = super().__deepcopy__(memo)
        obj.wrapper_attrs = self.wrapper_attrs.copy()
        return obj

    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        context['widget']['wrapper_attrs'] = self.wrapper_attrs
        return context



__all__ = [
    "PWOpenChoiceWidget",
]

from ._base import OpenChoiceWidget


class PWOpenChoiceWidget(OpenChoiceWidget):
    anchor_class = "PWProtectionTarget"

    class Media:
        overwrite = True



# Generated by Django 2.2.2 on 2019-06-21 20:12

from django.db import migrations


def move_keys(apps, schema_editor):
    AssignedContent = apps.get_model("spider_base", "AssignedContent")
    AnchorKey = apps.get_model("spider_keys", "AnchorKey")

    for i in AssignedContent.objects.filter(
        ctype__name="AnchorKey"
    ):
        content = AnchorKey.objects.get(id=i.object_id)
        i.attached_to_content = content.key
        i.save(update_fields=["attached_to_content"])


class Migration(migrations.Migration):

    dependencies = [
        ('spider_keys', '0002_20190426_squashed'),
    ]

    operations = [
        migrations.RunPython(move_keys),
    ]



__all__ = {
    "validate", "valid_wait_states", "verify_download_size",
    "async_validate", "verify_tag", "async_verify_tag"
}

import logging
import tempfile
import io

from django.utils.translation import gettext as _
from django.core.files import File
from django.conf import settings
from django.core import exceptions
from django.test import Client

from rdflib import Graph, URIRef, Literal
from rdflib.namespace import XSD
import requests

from spkcspider import celery_app

from spkcspider.apps.spider.constants import spkcgraph
from spkcspider.apps.spider.helpers import merge_get_url, get_settings_func

from .constants import BUFFER_SIZE
# uses specialized get_hashob from verifier (can be further customized)
from .functions import get_hashob
from .functions import get_requests_params, get_anchor_domain
from .models import VerifySourceObject, DataVerificationTag

logger = logging.getLogger(__name__)

hashable_predicates = set([spkcgraph["name"], spkcgraph["value"]])

valid_wait_states = {
    "RETRIEVING", "HASHING", "STARTED"
}


def verify_download_size(length, current_size=0):
    if not length or not length.isdigit():
        return False
    length = int(length)
    if settings.VERIFIER_MAX_SIZE_ACCEPTED < length:
        return False
    return True


def retrieve_object(obj, current_size, graph=None, session=None):
    # without graph return hash
    ret = None
    try:
        if not isinstance(obj, str):
            if graph is not None:
                graph.parse(obj, format="turtle")
            else:
                h = get_hashob()
                h.update(XSD.base64Binary.encode("utf8"))
                for chunk in iter(lambda: obj.read(BUFFER_SIZE), b''):
                    h.update(chunk)
                return h.finalize()

            return
        params, can_inline = get_requests_params(obj)
        if can_inline:
            resp = Client().get(obj)
            if resp.status_code != 200:
                raise exceptions.ValidationError(
                    _("Retrieval failed: %(reason)s"),
                    params={"reason": resp.reason},
                    code="error_code:{}".format(resp.status_code)
                )

            c_length = resp.get("content-length", None)
            if not verify_download_size(c_length, current_size[0]):
                resp.close()
                raise exceptions.ValidationError(
                    _("Content too big: %(size)s"),
                    params={"size": c_length},
                    code="invalid_size"
                )
            c_length = int(c_length)
            current_size[0] += c_length
            if graph is not None:
                graph.parse(getattr(
                    resp, "streaming_content", io.BytesIO(resp.content)
                ), format="turtle")
            else:
                h = get_hashob()
                h.update(XSD.base64Binary.encode("utf8"))
                for chunk in resp:
                    h.update(chunk)
                ret = h.finalize()
            resp.close()
        else:
            if not session:
                session = requests
            try:
                with session.get(
                    obj, stream=True, **params
                ) as resp:
                    if resp.status_code != 200:
                        raise exceptions.ValidationError(
                            _("Retrieval failed: %(reason)s"),
                            params={"reason": resp.reason},
                            code="error_code:{}".format(resp.status_code)
                        )

                    c_length = resp.headers.get("content-length", None)
                    if not verify_download_size(c_length, current_size[0]):
                        raise exceptions.ValidationError(
                            _("Content too big: %(size)s"),
                            params={"size": c_length},
                            code="invalid_size"
                        )
                    c_length = int(c_length)
                    current_size += c_length
                    if graph is not None:
                        graph.parse(resp.raw, format="turtle")
                    else:
                        h = get_hashob()
                        h.update(XSD.base64Binary.encode("utf8"))
                        for chunk in resp.iter_content(BUFFER_SIZE):
                            h.update(chunk)
                        ret = h.finalize()

            except requests.exceptions.Timeout:
                raise exceptions.ValidationError(
                    _('url timed out: %(url)s'),
                    params={"url": obj},
                    code="timeout_url"
                )
            except requests.exceptions.ConnectionError:
                raise exceptions.ValidationError(
                    _('invalid url: %(url)s'),
                    params={"url": obj},
                    code="invalid_url"
                )
    except Exception:
        logger.error("Parsing graph failed", exc_info=settings.DEBUG)
        raise exceptions.ValidationError(
            _('Invalid graph fromat'),
            code="invalid_format"
        )
    return ret


def validate(ob, hostpart, task=None):
    dvfile = None
    source = None
    g = Graph()
    g.namespace_manager.bind("spkc", spkcgraph, replace=True)
    with requests.session() as session:
        if isinstance(ob, tuple):
            current_size = ob[1]
            with open(ob[0], "rb") as f:
                retrieve_object(f, [current_size], graph=g, session=session)
        else:
            current_size = 0

            source = VerifySourceObject.objects.get(
                id=ob
            )
            url = source.get_url()
            retrieve_object(url, [current_size], graph=g, session=session)

        tmp = list(g.query(
            """
                SELECT ?base ?scope ?pages ?view
                WHERE {
                    ?base spkc:scope ?scope .
                    ?base spkc:pages.num_pages ?pages .
                    ?base spkc:pages.current_page ?current_page .
                    ?base spkc:action:view ?view .
                }
            """,
            initNs={"spkc": spkcgraph},
            initBindings={
                "current_page": Literal(1, datatype=XSD.positiveInteger)
            }
        ))

        if len(tmp) != 1:
            raise exceptions.ValidationError(
                _('Invalid graph'),
                code="invalid_graph"
            )
        tmp = tmp[0]
        start = tmp.base
        # scope = tmp.scope.toPython()
        view_url = tmp.view.toPython()
        pages = tmp.pages.toPython()

        if task:
            task.update_state(
                state='RETRIEVING',
                meta={
                    'page': 1,
                    'num_pages': pages
                }
            )
        if isinstance(ob, tuple):
            split = view_url.split("?", 1)
            source = VerifySourceObject.objects.update_or_create(
                url=split[0], defaults={"get_params": split[1]}
            )

        # retrieve further pages
        for page in range(2, pages+1):
            url = merge_get_url(
                source.get_url(), raw="embed", page=str(page)
            )
            retrieve_object(url, [current_size], graph=g, session=session)

            if task:
                task.update_state(
                    state='RETRIEVING',
                    meta={
                        'page': page,
                        'num_pages': pages
                    }
                )

        # check and clean graph
        data_type = get_settings_func(
            "VERIFIER_CLEAN_GRAPH",
            "spkcspider.apps.verifier.functions.clean_graph"
        )(g, start, source, hostpart)
        if not data_type:
            raise exceptions.ValidationError(
                _('Invalid graph (Verification failed)'),
                code="graph_failed"
            )
        g.remove((None, spkcgraph["csrftoken"], None))

        hashable_nodes = g.query(
            """
                SELECT DISTINCT ?base ?type ?name ?value
                WHERE {
                    ?base spkc:type ?type .
                    ?base spkc:properties ?prop .
                    ?prop spkc:hashable "true"^^xsd:boolean .
                    ?prop spkc:name ?name .
                    ?prop spkc:value ?value .
                    OPTIONAL {
                        ?value spkc:properties ?prop2 .
                        ?prop2 spkc:name "info"^^xsd:string .
                        ?prop2 spkc:value ?info .
                    } .
                }
            """,
            initNs={"spkc": spkcgraph}
        )

        if task:
            task.update_state(
                state='HASHING',
                meta={
                    'hashable_nodes_checked': 0
                }
            )
        # make sure triples are linked to start
        # (user can provide arbitary data)
        g.remove((start, spkcgraph["hashed"], None))
        # MAYBE: think about logic for incoperating hashes
        g.remove((start, spkcgraph["hash"], None))
        nodes = {}
        resources_with_hash = {}
        for count, val in enumerate(hashable_nodes, start=1):
            if isinstance(val.value, URIRef):
                assert(val.info)
                h = get_hashob()
                # should always hash with xsd.string
                h.update(XSD.string.encode("utf8"))
                # don't strip id, as some contents seperate only by id
                h.update(str(val.info).encode("utf8"))
                _hash = h.finalize()
            elif val.value.datatype == spkcgraph["hashableURI"]:
                _hash = resources_with_hash.get(val.value.value)
                if not _hash:
                    url = merge_get_url(val.value.value, raw="embed")
                    if not get_settings_func(
                        "SPIDER_URL_VALIDATOR",
                        "spkcspider.apps.spider.functions.validate_url_default"
                    )(url):
                        raise exceptions.ValidationError(
                            _('Insecure url: %(url)s'),
                            params={"url": url},
                            code="insecure_url"
                        )
                    _hash = retrieve_object(
                        url, [current_size], session=session
                    )
                    # do not use add as it could be corrupted by user
                    # (user can provide arbitary data)
                    _uri = URIRef(val.value.value)
                    g.set((
                        _uri,
                        spkcgraph["hash"],
                        Literal(_hash.hex())
                    ))
            else:
                h = get_hashob()
                if val.value.datatype == XSD.base64Binary:
                    h.update(val.value.datatype.encode("utf8"))
                    h.update(val.value.toPython())
                elif val.value.datatype:
                    h.update(val.value.datatype.encode("utf8"))
                    h.update(val.value.encode("utf8"))
                else:
                    h.update(XSD.string.encode("utf8"))
                    h.update(val.value.encode("utf8"))
                _hash = h.finalize()
            h = get_hashob()
            h.update(val.name.toPython().encode("utf8"))
            h.update(_hash)
            base = str(val.base)
            nodes.setdefault(base, ([], val.type))
            nodes[base][0].append(h.finalize())
            if task:
                task.update_state(
                    state='HASHING',
                    meta={
                        'hashable_nodes_checked': count
                    }
                )
    if task:
        task.update_state(
            state='HASHING',
            meta={
                'hashable_nodes_checked': "all"
            }
        )

    # first sort hashes per node and create hash over sorted hashes
    # de-duplicate super-hashes (means: nodes are identical)
    hashes = set()
    for val, _type in nodes.values():
        h = get_hashob()
        for _hob in sorted(val):
            h.update(_hob)
        h.update(_type.encode("utf8"))
        hashes.add(h.finalize())

    # then create hash over sorted de-duplicated node hashes
    h = get_hashob()
    for i in sorted(hashes):
        h.update(i)
    # do not use add as it could be corrupted by user
    # (user can provide arbitary data)
    digest = h.finalize().hex()
    g.set((
        start,
        spkcgraph["hash"],
        Literal(digest)
    ))

    with tempfile.NamedTemporaryFile(delete=True) as dvfile:
        # save in temporary file
        g.serialize(
            dvfile, format="turtle"
        )

        result, created = DataVerificationTag.objects.get_or_create(
            defaults={
                "dvfile": File(dvfile),
                "source": source,
                "data_type": data_type
            },
            hash=digest
        )
    update_fields = set()
    # and source, cannot remove source without replacement
    if not created and source and source != result.source:
        result.source = source
        update_fields.add("source")
    if data_type != result.data_type:
        result.data_type = data_type
        update_fields.add("data_type")
    result.save(update_fields=update_fields)
    verify_tag(result, task=task, ffrom="validate")
    if task:
        task.update_state(
            state='SUCCESS'
        )
    return result


@celery_app.task(bind=True, name='async validation')
def async_validate(self, ob, hostpart):
    ret = validate(ob, hostpart, self)
    return ret.get_absolute_url()


def verify_tag(tag, hostpart=None, ffrom="sync_call", task=None):
    """ for auto validation or hooks"""
    if not hostpart:
        hostpart = get_anchor_domain()
    if task:
        task.update_state(
            state='VERIFY'
        )

    if get_settings_func(
        "VERIFIER_TAG_VERIFIER",
        "spkcspider.apps.verifier.functions.verify_tag_default"
    )(tag, hostpart, ffrom):
        try:
            tag.callback(hostpart)
        except exceptions.ValidationError:
            logger.exception("Error while calling back")
    if task:
        task.update_state(
            state='SUCCESS'
        )


@celery_app.task(bind=True, name='async verification', ignore_results=True)
def async_verify_tag(self, tagid, hostpart=None, ffrom="async_call"):
    verify_tag(
        tag=DataVerificationTag.objects.get(id=tagid),
        hostpart=hostpart, task=self, ffrom=ffrom
    )



# Generated by Django 2.2.2 on 2019-06-21 20:30

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('spider_keys', '0003_auto_20190621_2012'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='anchorkey',
            name='key',
        ),
    ]



# Generated by Django 2.2.2 on 2019-06-21 21:26

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0003_auto_20190621_1954'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='linkcontent',
            name='content',
        ),
        migrations.AlterField(
            model_name='assignedcontent',
            name='content_type',
            field=models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType'),
        ),
    ]



__all__ = ["TokenUserMiddleware"]

from django.utils import timezone
from django.utils.functional import SimpleLazyObject
from .models import AuthToken


def get_user(request):
    # check for admin auth token
    token = request.headers.get("X-TOKEN", None)
    if token:
        now = timezone.now()

        token = AuthToken.objects.filter(
            usercomponent__name="index",
            token=token
        ).first()
        if token:
            uc = token.usercomponent
            if (
                # token.persist == -1 and  # cannot persist
                token.created < now - uc.token_duration
            ):
                token.delete()
                return None
            return uc.user
    return None


def get_cached_user(request):
    if not hasattr(request, "_cached_spidertoken_user"):
        request._cached_spidertoken_user = get_user(request)
    return request._cached_spidertoken_user


class TokenUserMiddleware(object):
    """ should come after AuthenticationMiddleware """
    def __init__(self, get_response=None):
        self.get_response = get_response
        super().__init__()

    def __call__(self, request):
        if hasattr(request, "user"):
            user = request.user
            request.user = SimpleLazyObject(
                lambda: get_cached_user(request) or user
            )
        else:
            request.user = SimpleLazyObject(
                lambda: get_cached_user(request)
            )
        return self.get_response(request)



from django import template
register = template.Library()


@register.filter(name='divide')
def divide(value, arg):
    return value//int(arg)


@register.filter(name='remainder')
def remainder(value, arg):
    return value % int(arg)



# Generated by Django 2.2 on 2019-04-19 08:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_filets', '0011_auto_20190405_0045'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='filefilet',
            name='license',
        ),
        migrations.RemoveField(
            model_name='textfilet',
            name='license',
        ),
        migrations.AddField(
            model_name='filefilet',
            name='license_url',
            field=models.URLField(blank=True, max_length=400, null=True),
        ),
        migrations.AddField(
            model_name='textfilet',
            name='license_url',
            field=models.URLField(blank=True, max_length=400, null=True),
        ),
    ]



__all__ = ["installed_fields", "generate_fields"]

import logging
import posixpath
from django import forms
from django.apps import apps
from django.utils.translation import gettext, gettext_lazy

from spkcspider.apps.spider.helpers import add_by_field

from spkcspider.apps.spider.widgets import (
    SubSectionStartWidget, SubSectionStopWidget
)

installed_fields = {}

safe_default_fields = [
    "BooleanField", "CharField", "ChoiceField", "MultipleChoiceField",
    "DateField", "DateTimeField", "DecimalField", "DurationField",
    "EmailField", "FilePathField", "FloatField", "GenericIPAddressField",
    "ModelChoiceField", "ModelMultipleChoiceField", "SlugField", "TimeField",
    "URLField"
]
for i in safe_default_fields:
    installed_fields[i] = getattr(forms, i)


@add_by_field(installed_fields, "__name__")
class TextareaField(forms.CharField):
    widget = forms.Textarea

# extra attributes for fields:
# limit_to_usercomponent = "<fieldname">: limit field name to associated uc
# limit_to_user = "<fieldname">: limit field name to user of associated uc


def localized_choices(obj):
    def func(*, choices=(), **kwargs):
        choices = map(lambda x: (x[0], gettext(x[1])), choices)
        return obj(choices=choices, **kwargs)
    return func


installed_fields["LocalizedChoiceField"] = localized_choices(forms.ChoiceField)
installed_fields["MultipleLocalizedChoiceField"] = \
    localized_choices(forms.MultipleChoiceField)


@add_by_field(installed_fields, "__name__")
class UserContentRefField(forms.ModelChoiceField):
    filter_strength_link = "associated_rel__usercomponent__strength__lte"
    exclude_travel = (
        "associated_rel__usercomponent__travel_protected__in",
        "associated_rel__travel_protected__in",
    )

    # limit_to_uc: limit to usercomponent, if False to user
    # True is strongly recommended to prevent info leak gadgets
    def __init__(self, modelname, limit_to_uc=True, **kwargs):
        from spkcspider.apps.spider.contents import BaseContent
        if limit_to_uc:
            self.filters_usercomponent = (
                "associated_rel__usercomponent",
                "associated_rel__referenced_by__usercomponent"
            )
        else:
            self.filters_user = ("associated_rel__usercomponent__user",)

        model = apps.get_model(modelname)
        if not issubclass(model, BaseContent):
            raise Exception("Not a content (inherit from BaseContent)")

        kwargs["queryset"] = model.objects.filter(
            **kwargs.pop("limit_choices_to", {})
        )
        super().__init__(**kwargs)

    def tagdata_from_value(self, obj):
        if obj:
            return obj.pk
        return None

    def label_from_instance(self, obj):
        return str(obj)


@add_by_field(installed_fields, "__name__")
class MultipleUserContentRefField(forms.ModelMultipleChoiceField):
    filter_strength_link = "associated_rel__usercomponent__strength__lte"
    exclude_travel = (
        "associated_rel__usercomponent__travel_protected__in",
        "associated_rel__travel_protected__in",
    )

    # limit_to_uc: limit to usercomponent, if False to user
    # True is strongly recommended to prevent info leak gadgets
    def __init__(self, modelname, limit_to_uc=True, **kwargs):
        from spkcspider.apps.spider.contents import BaseContent
        if limit_to_uc:
            self.filters_usercomponent = (
                "associated_rel__usercomponent",
                "associated_rel__referenced_by__usercomponent"
            )
        else:
            self.filters_user = ("associated_rel__usercomponent__user",)

        model = apps.get_model(
            modelname
        )
        if not issubclass(model, BaseContent):
            raise Exception("Not a content (inherit from BaseContent)")

        kwargs["queryset"] = model.objects.filter(
            **kwargs.pop("limit_choices_to", {})
        )
        super().__init__(**kwargs)

    def tagdata_from_value(self, query):
        return list(query.values_list(
            'pk', flat=True
        ))

    def label_from_instance(self, obj):
        return str(obj)


@add_by_field(installed_fields, "__name__")
class AnchorField(forms.ModelChoiceField):
    spkc_use_uriref = True
    use_default_anchor = None
    filter_strength_link = "usercomponent__strength__lte"
    exclude_travel = (
        "associated_rel__usercomponent__travel_protected__in",
        "associated_rel__travel_protected__in",
    )

    # limit_to_uc: limit to usercomponent, if False to user
    def __init__(
        self, use_default_anchor=True, limit_to_uc=True, **kwargs
    ):
        from spkcspider.apps.spider.models import AssignedContent
        _ = gettext_lazy
        if limit_to_uc:
            self.filters_usercomponent = ("usercomponent",)
        else:
            self.filters_user = ("usercomponent__user",)
        self.use_default_anchor = use_default_anchor
        if use_default_anchor:
            kwargs.setdefault("required", False)
        if use_default_anchor and not kwargs.get("empty_label", None):
            kwargs["empty_label"] = _("(Use default anchor)")

        # can also use Links to anchor as anchor
        kwargs["queryset"] = AssignedContent.objects.filter(
            info__contains="\x1eanchor\x1e",
            **kwargs.pop("limit_choices_to", {})
        )
        super().__init__(**kwargs)

    def to_python(self, value):
        if value in self.empty_values:
            return None
        try:
            key = self.to_field_name or 'pk'
            value = self.queryset.get(**{key: value}).content
        except (ValueError, TypeError, self.queryset.model.DoesNotExist):
            raise forms.ValidationError(
                self.error_messages['invalid_choice'], code='invalid_choice'
            )
        return value

    def tagdata_from_value(self, obj):
        if obj:
            return obj.associated.id
        return None

    def label_from_instance(self, obj):
        return str(obj.content)


class StartSub(forms.Field):
    widget = SubSectionStartWidget
    hashable = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.widget.label = self.label


class StopSub(forms.Field):
    widget = SubSectionStopWidget
    hashable = False


def generate_fields(layout, prefix="", _base=None, _mainprefix=None):
    if _base is None:
        _base = []
        _mainprefix = prefix
    for i in layout:
        item = i.copy()
        item.setdefault("required", False)
        try:
            key, field = item.pop("key", None), item.pop("field", None)
        except Exception:
            logging.warning("Invalid item (no dict)", i)
            continue
        localize = item.pop("localize", False)
        nonhashable = item.pop("nonhashable", False)
        if not item.get("label"):
            item["label"] = key.replace(_mainprefix, "", 1)

        if localize:
            item["label"] = gettext(item["label"])
            if "help_text" in item:
                item["help_text"] = gettext(item["help_text"])
        if not key or "/" in key:
            logging.warning("Invalid item (no key/contains /)", i)
            continue
        if isinstance(field, list):
            new_prefix = posixpath.join(prefix, key)
            item["required"] = False
            item["initial"] = None
            # by prefixing with _ invalidate prefix for tag recognition
            _base.append((
                "_{}_start".format(new_prefix),
                StartSub(**item)
            ))
            generate_fields(
                field, new_prefix, _base=_base, _mainprefix=_mainprefix
            )
            # by prefixing with _ invalidate prefix for tag recognition
            _base.append((
                "_{}_stop".format(new_prefix),
                StopSub(**item)
            ))
        elif isinstance(field, str):
            new_field = installed_fields.get(field, None)
            if not new_field:
                logging.warning("Invalid field specified: %s", field)
            else:
                new_field = new_field(**item)
                setattr(new_field, "hashable", not nonhashable)
                _base.append((posixpath.join(prefix, key), new_field))
        else:
            logging.warning("Invalid item", i)
    return _base



from django import template
from django import forms
from django.template.loader import render_to_string
from ..protections import ProtectionList

register = template.Library()


@register.simple_tag(takes_context=True)
def render_protection(context, protectiontup):
    result, protection = protectiontup
    protection = protection.installed_class
    ctx = {}
    ctx["parent_ctx"] = context
    if "hostpart" in context:
        ctx["hostpart"] = context["hostpart"]
    ctx["data"] = result
    ctx["protection"] = protection

    if callable(getattr(protection, "render", None)):
        return protection.render(ctx)
    if isinstance(ctx["data"], forms.Form):
        ctx["form"] = ctx["data"]
    elif hasattr(protection, "auth_form") and isinstance(ctx["data"], int):
        ctx["form"] = protection.auth_form()
    elif hasattr(protection, "auth_form"):
        ctx["form"] = protection.auth_form(**ctx["data"])
    template_name = getattr(protection, "template_name", None)
    if not template_name:
        template_name = "spider_base/protections/protection_form.html"

    return render_to_string(
        template_name, context=ctx, request=context["request"]
    )


@register.simple_tag(takes_context=True)
def extract_protections(context, extract_name="protections"):
    if isinstance(
        getattr(
            context["request"], extract_name, None
        ), ProtectionList
    ):
        return getattr(context["request"], extract_name)
    return ProtectionList()
